# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2014 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************


from PySide import QtCore
from PySide.QtCore import QT_TRANSLATE_NOOP
import FreeCAD
import Path
import Path.Dressup.Utils as PathDressup
import PathScripts.PathUtils as PathUtils
import math
from pivy import coin

# lazily loaded modules
from lazy_loader.lazy_loader import LazyLoader

DraftGeomUtils = LazyLoader("DraftGeomUtils", globals(), "DraftGeomUtils")
Part = LazyLoader("Part", globals(), "Part")


if False:
    Path.Log.setLevel(Path.Log.Level.DEBUG, Path.Log.thisModule())
    Path.Log.trackModule(Path.Log.thisModule())
else:
    Path.Log.setLevel(Path.Log.Level.INFO, Path.Log.thisModule())


translate = FreeCAD.Qt.translate


movecommands = Path.Geom.CmdMoveStraight + Path.Geom.CmdMoveRapid + Path.Geom.CmdMoveArc


def debugMarker(vector, label, color=None, radius=0.5):
    if Path.Log.getLevel(Path.Log.thisModule()) == Path.Log.Level.DEBUG:
        obj = FreeCAD.ActiveDocument.addObject("Part::Sphere", label)
        obj.Label = label
        obj.Radius = radius
        obj.Placement = FreeCAD.Placement(vector, FreeCAD.Rotation(FreeCAD.Vector(0, 0, 1), 0))
        if color:
            obj.ViewObject.ShapeColor = color


def debugCircle(vector, r, label, color=None):
    if Path.Log.getLevel(Path.Log.thisModule()) == Path.Log.Level.DEBUG:
        obj = FreeCAD.ActiveDocument.addObject("Part::Cylinder", label)
        obj.Label = label
        obj.Radius = r
        obj.Height = 1
        obj.Placement = FreeCAD.Placement(vector, FreeCAD.Rotation(FreeCAD.Vector(0, 0, 1), 0))
        obj.ViewObject.Transparency = 90
        if color:
            obj.ViewObject.ShapeColor = color


def addAngle(a1, a2):
    a = a1 + a2
    while a <= -math.pi:
        a += 2 * math.pi
    while a > math.pi:
        a -= 2 * math.pi
    return a


def anglesAreParallel(a1, a2):
    an1 = addAngle(a1, 0)
    an2 = addAngle(a2, 0)
    if an1 == an2:
        return True
    if an1 == addAngle(an2, math.pi):
        return True
    return False


def getAngle(v):
    a = v.getAngle(FreeCAD.Vector(1, 0, 0))
    if v.y < 0:
        return -a
    return a


def pointFromCommand(cmd, pt, X="X", Y="Y", Z="Z"):
    x = cmd.Parameters.get(X, pt.x)
    y = cmd.Parameters.get(Y, pt.y)
    z = cmd.Parameters.get(Z, pt.z)
    return FreeCAD.Vector(x, y, z)


def edgesForCommands(cmds, startPt):
    edges = []
    lastPt = startPt
    for cmd in cmds:
        if cmd.Name in movecommands:
            pt = pointFromCommand(cmd, lastPt)
            if cmd.Name in Path.Geom.CmdMoveStraight:
                edges.append(Part.Edge(Part.LineSegment(lastPt, pt)))
            elif cmd.Name in Path.Geom.CmdMoveArc:
                center = lastPt + pointFromCommand(cmd, FreeCAD.Vector(0, 0, 0), "I", "J", "K")
                A = lastPt - center
                B = pt - center
                d = -B.x * A.y + B.y * A.x

                if d == 0:
                    # we're dealing with half a circle here
                    angle = getAngle(A) + math.pi / 2
                    if cmd.Name in Path.Geom.CmdMoveCW:
                        angle -= math.pi
                else:
                    C = A + B
                    angle = getAngle(C)

                R = (lastPt - center).Length
                ptm = center + FreeCAD.Vector(math.cos(angle), math.sin(angle), 0) * R

                edges.append(Part.Edge(Part.Arc(lastPt, ptm, pt)))
            lastPt = pt
    return edges


class Style(object):

    Dogbone = "Dogbone"
    Tbone_H = "T-bone horizontal"
    Tbone_V = "T-bone vertical"
    Tbone_L = "T-bone long edge"
    Tbone_S = "T-bone short edge"
    All = [Dogbone, Tbone_H, Tbone_V, Tbone_L, Tbone_S]


class Side(object):

    Left = "Left"
    Right = "Right"
    All = [Left, Right]

    @classmethod
    def oppositeOf(cls, side):
        if side == cls.Left:
            return cls.Right
        if side == cls.Right:
            return cls.Left
        return None


class Incision(object):

    Fixed = "fixed"
    Adaptive = "adaptive"
    Custom = "custom"
    All = [Adaptive, Fixed, Custom]


class Smooth(object):

    Neither = 0
    In = 1
    Out = 2
    InAndOut = In | Out


# Chord
# A class to represent the start and end point of a path command. If the underlying
# Command is a rotate command the receiver does represent a chord in the geometric
# sense of the word. If the underlying command is a straight move then the receiver
# represents the actual move.
# This implementation really only deals with paths in the XY plane. Z is assumed to
# be constant in all calculated results.
# Instances of Chord are generally considered immutable and all movement member
# functions return new instances.
class Chord(object):
    def __init__(self, start=None, end=None):
        if not start:
            start = FreeCAD.Vector()
        if not end:
            end = FreeCAD.Vector()
        self.Start = start
        self.End = end

    def __str__(self):
        return "Chord([%g, %g, %g] -> [%g, %g, %g])" % (
            self.Start.x,
            self.Start.y,
            self.Start.z,
            self.End.x,
            self.End.y,
            self.End.z,
        )

    def moveTo(self, newEnd):
        return Chord(self.End, newEnd)

    def moveToParameters(self, params):
        x = params.get("X", self.End.x)
        y = params.get("Y", self.End.y)
        z = params.get("Z", self.End.z)
        return self.moveTo(FreeCAD.Vector(x, y, z))

    def moveBy(self, x, y, z):
        return self.moveTo(self.End + FreeCAD.Vector(x, y, z))

    def move(self, distance, angle):
        dx = distance * math.cos(angle)
        dy = distance * math.sin(angle)
        return self.moveBy(dx, dy, 0)

    def asVector(self):
        return self.End - self.Start

    def asDirection(self):
        return self.asVector().normalize()

    def asLine(self):
        return Part.LineSegment(self.Start, self.End)

    def asEdge(self):
        return Part.Edge(self.asLine())

    def getLength(self):
        return self.asVector().Length

    def getDirectionOfVector(self, B):
        A = self.asDirection()
        # if the 2 vectors are identical, they head in the same direction
        Path.Log.debug("   {}.getDirectionOfVector({})".format(A, B))
        if Path.Geom.pointsCoincide(A, B):
            return "Straight"
        d = -A.x * B.y + A.y * B.x
        if d < 0:
            return Side.Left
        if d > 0:
            return Side.Right
        # at this point the only direction left is backwards
        return "Back"

    def getDirectionOf(self, chordOrVector):
        if type(chordOrVector) is Chord:
            return self.getDirectionOfVector(chordOrVector.asDirection())
        return self.getDirectionOfVector(chordOrVector.normalize())

    def getAngleOfVector(self, ref):
        angle = self.asVector().getAngle(ref)
        # unfortunately they never figure out the sign :(
        # positive angles go up, so when the reference vector is left
        # then the receiver must go down
        if self.getDirectionOfVector(ref) == Side.Left:
            return -angle
        return angle

    def getAngle(self, refChordOrVector):
        if type(refChordOrVector) is Chord:
            return self.getAngleOfVector(refChordOrVector.asDirection())
        return self.getAngleOfVector(refChordOrVector.normalize())

    def getAngleXY(self):
        return self.getAngle(FreeCAD.Vector(1, 0, 0))

    def commandParams(self, f):
        params = {"X": self.End.x, "Y": self.End.y, "Z": self.End.z}
        if f:
            params["F"] = f
        return params

    def g1Command(self, f):
        return Path.Command("G1", self.commandParams(f))

    def arcCommand(self, cmd, center, f):
        params = self.commandParams(f)
        d = center - self.Start
        params["I"] = d.x
        params["J"] = d.y
        params["K"] = 0
        return Path.Command(cmd, params)

    def g2Command(self, center, f):
        return self.arcCommand("G2", center, f)

    def g3Command(self, center, f):
        return self.arcCommand("G3", center, f)

    def isAPlungeMove(self):
        return not Path.Geom.isRoughly(self.End.z, self.Start.z)

    def isANoopMove(self):
        Path.Log.debug(
            "{}.isANoopMove(): {}".format(self, Path.Geom.pointsCoincide(self.Start, self.End))
        )
        return Path.Geom.pointsCoincide(self.Start, self.End)

    def foldsBackOrTurns(self, chord, side):
        direction = chord.getDirectionOf(self)
        Path.Log.info("  - direction = %s/%s" % (direction, side))
        return direction == "Back" or direction == side

    def connectsTo(self, chord):
        return Path.Geom.pointsCoincide(self.End, chord.Start)


class Bone(object):
    def __init__(self, boneId, obj, lastCommand, inChord, outChord, smooth, F):
        self.obj = obj
        self.boneId = boneId
        self.lastCommand = lastCommand
        self.inChord = inChord
        self.outChord = outChord
        self.smooth = smooth
        self.smooth = Smooth.Neither
        self.F = F

        # initialized later
        self.cDist = None
        self.cAngle = None
        self.tAngle = None
        self.cPt = None

    def angle(self):
        if self.cAngle is None:
            baseAngle = self.inChord.getAngleXY()
            turnAngle = self.outChord.getAngle(self.inChord)
            theta = addAngle(baseAngle, (turnAngle - math.pi) / 2)
            if self.obj.Side == Side.Left:
                theta = addAngle(theta, math.pi)
            self.tAngle = turnAngle
            self.cAngle = theta
        return self.cAngle

    def distance(self, toolRadius):
        if self.cDist is None:
            self.angle()  # make sure the angles are initialized
            self.cDist = toolRadius / math.cos(self.tAngle / 2)
        return self.cDist

    def corner(self, toolRadius):
        if self.cPt is None:
            self.cPt = self.inChord.move(self.distance(toolRadius), self.angle()).End
        return self.cPt

    def location(self):
        return (self.inChord.End.x, self.inChord.End.y)

    def locationZ(self):
        return (self.inChord.End.x, self.inChord.End.y, self.inChord.End.z)

    def adaptiveLength(self, boneAngle, toolRadius):
        theta = self.angle()
        distance = self.distance(toolRadius)
        # there is something weird happening if the boneAngle came from a horizontal/vertical t-bone
        # for some reason pi/2 is not equal to pi/2
        if math.fabs(theta - boneAngle) < 0.00001:
            # moving directly towards the corner
            Path.Log.debug("adaptive - on target: %.2f - %.2f" % (distance, toolRadius))
            return distance - toolRadius
        Path.Log.debug(
            "adaptive - angles: corner=%.2f  bone=%.2f diff=%.12f"
            % (theta / math.pi, boneAngle / math.pi, theta - boneAngle)
        )

        # The bones root and end point form a triangle with the intersection of the tool path
        # with the toolRadius circle around the bone end point.
        # In case the math looks questionable, look for "triangle ssa"
        # c = distance
        # b = self.toolRadius
        # beta = fabs(boneAngle - theta)
        beta = math.fabs(addAngle(boneAngle, -theta))
        D = (distance / toolRadius) * math.sin(beta)
        if D > 1:  # no intersection
            Path.Log.debug("adaptive - no intersection - no bone")
            return 0
        gamma = math.asin(D)
        alpha = math.pi - beta - gamma
        if Path.Geom.isRoughly(0.0, math.sin(beta)):
            # it is not a good idea to divide by 0
            length = 0.0
        else:
            length = toolRadius * math.sin(alpha) / math.sin(beta)
            if D < 1 and toolRadius < distance:  # there exists a second solution
                beta2 = beta
                gamma2 = math.pi - gamma
                alpha2 = math.pi - beta2 - gamma2
                length2 = toolRadius * math.sin(alpha2) / math.sin(beta2)
                length = min(length, length2)

        Path.Log.debug(
            "adaptive corner=%.2f * %.2f˚ -> bone=%.2f * %.2f˚"
            % (distance, theta, length, boneAngle)
        )
        return length


class ObjectDressup(object):
    def __init__(self, obj, base):
        # Tool Properties
        obj.addProperty(
            "App::PropertyLink",
            "Base",
            "Base",
            QT_TRANSLATE_NOOP("App::Property", "The base path to modify"),
        )
        obj.addProperty(
            "App::PropertyEnumeration",
            "Side",
            "Dressup",
            QT_TRANSLATE_NOOP("App::Property", "The side of path to insert bones"),
        )
        obj.Side = [Side.Left, Side.Right]
        obj.Side = Side.Right
        obj.addProperty(
            "App::PropertyEnumeration",
            "Style",
            "Dressup",
            QT_TRANSLATE_NOOP("App::Property", "The style of bones"),
        )
        obj.Style = Style.All
        obj.Style = Style.Dogbone
        obj.addProperty(
            "App::PropertyIntegerList",
            "BoneBlacklist",
            "Dressup",
            QT_TRANSLATE_NOOP("App::Property", "Bones that are not dressed up"),
        )
        obj.BoneBlacklist = []
        obj.setEditorMode("BoneBlacklist", 2)  # hide this one
        obj.addProperty(
            "App::PropertyEnumeration",
            "Incision",
            "Dressup",
            QT_TRANSLATE_NOOP("App::Property", "The algorithm to determine the bone length"),
        )
        obj.Incision = Incision.All
        obj.Incision = Incision.Adaptive
        obj.addProperty(
            "App::PropertyFloat",
            "Custom",
            "Dressup",
            QT_TRANSLATE_NOOP("App::Property", "Dressup length if incision is set to 'custom'"),
        )
        obj.Custom = 0.0
        obj.Proxy = self
        obj.Base = base

        # initialized later
        self.boneShapes = None
        self.toolRadius = 0
        self.dbg = None
        self.locationBlacklist = None
        self.shapes = None
        self.boneId = None
        self.bones = None

    def onDocumentRestored(self, obj):
        obj.setEditorMode("BoneBlacklist", 2)  # hide this one

    def dumps(self):
        return None

    def loads(self, state):
        return None

    def theOtherSideOf(self, side):
        if side == Side.Left:
            return Side.Right
        return Side.Left

    # Answer true if a dogbone could be on either end of the chord, given its command
    def canAttachDogbone(self, cmd, chord):
        return (
            cmd.Name in Path.Geom.CmdMoveStraight
            and not chord.isAPlungeMove()
            and not chord.isANoopMove()
        )

    def shouldInsertDogbone(self, obj, inChord, outChord):
        return outChord.foldsBackOrTurns(inChord, self.theOtherSideOf(obj.Side))

    def findPivotIntersection(self, pivot, pivotEdge, edge, refPt, d, color):
        Path.Log.track(
            "(%.2f, %.2f)^%.2f  - [(%.2f, %.2f), (%.2f, %.2f)]"
            % (
                pivotEdge.Curve.Center.x,
                pivotEdge.Curve.Center.y,
                pivotEdge.Curve.Radius,
                edge.Vertexes[0].Point.x,
                edge.Vertexes[0].Point.y,
                edge.Vertexes[1].Point.x,
                edge.Vertexes[1].Point.y,
            )
        )
        ppt = None
        pptDistance = 0
        for pt in DraftGeomUtils.findIntersection(edge, pivotEdge, dts=False):
            # debugMarker(pt, "pti.%d-%s.in" % (self.boneId, d), color, 0.2)
            distance = (pt - refPt).Length
            Path.Log.debug("        -->  (%.2f, %.2f): %.2f" % (pt.x, pt.y, distance))
            if not ppt or pptDistance < distance:
                ppt = pt
                pptDistance = distance
        if not ppt:
            tangent = DraftGeomUtils.findDistance(pivot, edge)
            if tangent:
                Path.Log.debug("Taking tangent as intersect %s" % tangent)
                ppt = pivot + tangent
            else:
                Path.Log.debug("Taking chord start as intersect %s" % edge.Vertexes[0].Point)
                ppt = edge.Vertexes[0].Point
            # debugMarker(ppt, "ptt.%d-%s.in" % (self.boneId, d), color, 0.2)
            Path.Log.debug("        -->  (%.2f, %.2f)" % (ppt.x, ppt.y))
        return ppt

    def pointIsOnEdge(self, point, edge):
        param = edge.Curve.parameter(point)
        return edge.FirstParameter <= param <= edge.LastParameter

    def smoothChordCommands(self, bone, inChord, outChord, edge, wire, corner, smooth, color=None):
        if smooth == 0:
            Path.Log.info(" No smoothing requested")
            return [bone.lastCommand, outChord.g1Command(bone.F)]

        d = "in"
        refPoint = inChord.Start
        if smooth == Smooth.Out:
            d = "out"
            refPoint = outChord.End

        if DraftGeomUtils.areColinear(inChord.asEdge(), outChord.asEdge()):
            Path.Log.info(" straight edge %s" % d)
            return [outChord.g1Command(bone.F)]

        pivot = None
        pivotDistance = 0

        Path.Log.info(
            "smooth:  (%.2f, %.2f)-(%.2f, %.2f)"
            % (
                edge.Vertexes[0].Point.x,
                edge.Vertexes[0].Point.y,
                edge.Vertexes[1].Point.x,
                edge.Vertexes[1].Point.y,
            )
        )
        for e in wire.Edges:
            self.dbg.append(e)
            if type(e.Curve) == Part.LineSegment or type(e.Curve) == Part.Line:
                Path.Log.debug(
                    "         (%.2f, %.2f)-(%.2f, %.2f)"
                    % (
                        e.Vertexes[0].Point.x,
                        e.Vertexes[0].Point.y,
                        e.Vertexes[1].Point.x,
                        e.Vertexes[1].Point.y,
                    )
                )
            else:
                Path.Log.debug(
                    "         (%.2f, %.2f)^%.2f"
                    % (e.Curve.Center.x, e.Curve.Center.y, e.Curve.Radius)
                )
            for pt in DraftGeomUtils.findIntersection(edge, e, True, findAll=True):
                if not Path.Geom.pointsCoincide(pt, corner) and self.pointIsOnEdge(pt, e):
                    # debugMarker(pt, "candidate-%d-%s" % (self.boneId, d), color, 0.05)
                    Path.Log.debug("         -> candidate")
                    distance = (pt - refPoint).Length
                    if not pivot or pivotDistance > distance:
                        pivot = pt
                        pivotDistance = distance
                else:
                    Path.Log.debug("         -> corner intersect")

        if pivot:
            # debugCircle(pivot, self.toolRadius, "pivot.%d-%s" % (self.boneId, d), color)

            pivotEdge = Part.Edge(Part.Circle(pivot, FreeCAD.Vector(0, 0, 1), self.toolRadius))
            t1 = self.findPivotIntersection(
                pivot, pivotEdge, inChord.asEdge(), inChord.End, d, color
            )
            t2 = self.findPivotIntersection(
                pivot, pivotEdge, outChord.asEdge(), inChord.End, d, color
            )

            commands = []
            if not Path.Geom.pointsCoincide(t1, inChord.Start):
                Path.Log.debug("  add lead in")
                commands.append(Chord(inChord.Start, t1).g1Command(bone.F))
            if bone.obj.Side == Side.Left:
                Path.Log.debug("  add g3 command")
                commands.append(Chord(t1, t2).g3Command(pivot, bone.F))
            else:
                Path.Log.debug(
                    "  add g2 command center=(%.2f, %.2f) -> from (%2f, %.2f) to (%.2f, %.2f"
                    % (pivot.x, pivot.y, t1.x, t1.y, t2.x, t2.y)
                )
                commands.append(Chord(t1, t2).g2Command(pivot, bone.F))
            if not Path.Geom.pointsCoincide(t2, outChord.End):
                Path.Log.debug("  add lead out")
                commands.append(Chord(t2, outChord.End).g1Command(bone.F))

            # debugMarker(pivot, "pivot.%d-%s"     % (self.boneId, d), color, 0.2)
            # debugMarker(t1,    "pivot.%d-%s.in"  % (self.boneId, d), color, 0.1)
            # debugMarker(t2,    "pivot.%d-%s.out" % (self.boneId, d), color, 0.1)

            return commands

        Path.Log.info(" no pivot found - straight command")
        return [inChord.g1Command(bone.F), outChord.g1Command(bone.F)]

    def inOutBoneCommands(self, bone, boneAngle, fixedLength):
        corner = bone.corner(self.toolRadius)

        bone.tip = bone.inChord.End  # in case there is no bone

        Path.Log.debug("corner = (%.2f, %.2f)" % (corner.x, corner.y))
        # debugMarker(corner, 'corner', (1., 0., 1.), self.toolRadius)

        length = fixedLength
        if bone.obj.Incision == Incision.Custom:
            length = bone.obj.Custom
        if bone.obj.Incision == Incision.Adaptive:
            length = bone.adaptiveLength(boneAngle, self.toolRadius)

        if length == 0:
            Path.Log.info("no bone after all ..")
            return [bone.lastCommand, bone.outChord.g1Command(bone.F)]

        # track length for marker visuals
        self.length = max(self.length, length)

        boneInChord = bone.inChord.move(length, boneAngle)
        boneOutChord = boneInChord.moveTo(bone.outChord.Start)

        # debugCircle(boneInChord.Start, self.toolRadius, 'boneStart')
        # debugCircle(boneInChord.End, self.toolRadius, 'boneEnd')

        bone.tip = boneInChord.End

        if bone.smooth == 0:
            return [
                bone.lastCommand,
                boneInChord.g1Command(bone.F),
                boneOutChord.g1Command(bone.F),
                bone.outChord.g1Command(bone.F),
            ]

        # reconstruct the corner and convert to an edge
        offset = corner - bone.inChord.End
        iChord = Chord(bone.inChord.Start + offset, bone.inChord.End + offset)
        oChord = Chord(bone.outChord.Start + offset, bone.outChord.End + offset)
        iLine = iChord.asLine()
        oLine = oChord.asLine()
        cornerShape = Part.Shape([iLine, oLine])

        # construct a shape representing the cut made by the bone
        vt0 = FreeCAD.Vector(0, self.toolRadius, 0)
        vt1 = FreeCAD.Vector(length, self.toolRadius, 0)
        vb0 = FreeCAD.Vector(0, -self.toolRadius, 0)
        vb1 = FreeCAD.Vector(length, -self.toolRadius, 0)
        vm2 = FreeCAD.Vector(length + self.toolRadius, 0, 0)

        boneBot = Part.LineSegment(vb1, vb0)
        boneLid = Part.LineSegment(vb0, vt0)
        boneTop = Part.LineSegment(vt0, vt1)

        # what we actually want is an Arc - but findIntersect only returns the coincident if one exists
        # which really sucks because that's the one we're probably not interested in ....
        boneArc = Part.Arc(vt1, vm2, vb1)
        # boneArc = Part.Circle(FreeCAD.Vector(length, 0, 0), FreeCAD.Vector(0,0,1), self.toolRadius)
        boneWire = Part.Shape([boneTop, boneArc, boneBot, boneLid])
        boneWire.rotate(FreeCAD.Vector(0, 0, 0), FreeCAD.Vector(0, 0, 1), boneAngle * 180 / math.pi)
        boneWire.translate(bone.inChord.End)
        self.boneShapes = [cornerShape, boneWire]

        bone.inCommands = self.smoothChordCommands(
            bone,
            bone.inChord,
            boneInChord,
            Part.Edge(iLine),
            boneWire,
            corner,
            bone.smooth & Smooth.In,
            (1.0, 0.0, 0.0),
        )
        bone.outCommands = self.smoothChordCommands(
            bone,
            boneOutChord,
            bone.outChord,
            Part.Edge(oLine),
            boneWire,
            corner,
            bone.smooth & Smooth.Out,
            (0.0, 1.0, 0.0),
        )
        return bone.inCommands + bone.outCommands

    def dogbone(self, bone):
        boneAngle = bone.angle()
        length = self.toolRadius * 0.41422  # 0.41422 = 2/sqrt(2) - 1 + (a tiny bit)
        return self.inOutBoneCommands(bone, boneAngle, length)

    def tboneHorizontal(self, bone):
        angle = bone.angle()
        boneAngle = 0
        if math.fabs(angle) > math.pi / 2:
            boneAngle = math.pi
        return self.inOutBoneCommands(bone, boneAngle, self.toolRadius)

    def tboneVertical(self, bone):
        angle = bone.angle()
        boneAngle = math.pi / 2
        if Path.Geom.isRoughly(angle, math.pi) or angle < 0:
            boneAngle = -boneAngle
        return self.inOutBoneCommands(bone, boneAngle, self.toolRadius)

    def tboneEdgeCommands(self, bone, onIn):
        if onIn:
            boneAngle = bone.inChord.getAngleXY()
        else:
            boneAngle = bone.outChord.getAngleXY()

        if Side.Right == bone.outChord.getDirectionOf(bone.inChord):
            boneAngle = boneAngle - math.pi / 2
        else:
            boneAngle = boneAngle + math.pi / 2

        onInString = "out"
        if onIn:
            onInString = "in"
        Path.Log.debug(
            "tboneEdge boneAngle[%s]=%.2f   (in=%.2f, out=%.2f)"
            % (
                onInString,
                boneAngle / math.pi,
                bone.inChord.getAngleXY() / math.pi,
                bone.outChord.getAngleXY() / math.pi,
            )
        )
        return self.inOutBoneCommands(bone, boneAngle, self.toolRadius)

    def tboneLongEdge(self, bone):
        inChordIsLonger = bone.inChord.getLength() > bone.outChord.getLength()
        return self.tboneEdgeCommands(bone, inChordIsLonger)

    def tboneShortEdge(self, bone):
        inChordIsShorter = bone.inChord.getLength() < bone.outChord.getLength()
        return self.tboneEdgeCommands(bone, inChordIsShorter)

    def boneIsBlacklisted(self, bone):
        blacklisted = False
        parentConsumed = False
        if bone.boneId in bone.obj.BoneBlacklist:
            blacklisted = True
        elif bone.location() in self.locationBlacklist:
            bone.obj.BoneBlacklist.append(bone.boneId)
            blacklisted = True
        elif hasattr(bone.obj.Base, "BoneBlacklist"):
            parentConsumed = bone.boneId not in bone.obj.Base.BoneBlacklist
            blacklisted = parentConsumed
        if blacklisted:
            self.locationBlacklist.add(bone.location())
        return (blacklisted, parentConsumed)

    # Generate commands necessary to execute the dogbone
    def boneCommands(self, bone, enabled):
        if enabled:
            if bone.obj.Style == Style.Dogbone:
                return self.dogbone(bone)
            if bone.obj.Style == Style.Tbone_H:
                return self.tboneHorizontal(bone)
            if bone.obj.Style == Style.Tbone_V:
                return self.tboneVertical(bone)
            if bone.obj.Style == Style.Tbone_L:
                return self.tboneLongEdge(bone)
            if bone.obj.Style == Style.Tbone_S:
                return self.tboneShortEdge(bone)
        else:
            return [bone.lastCommand, bone.outChord.g1Command(bone.F)]

    def insertBone(self, bone):
        Path.Log.debug(
            ">----------------------------------- %d --------------------------------------"
            % bone.boneId
        )
        self.boneShapes = []
        blacklisted, inaccessible = self.boneIsBlacklisted(bone)
        enabled = not blacklisted
        self.bones.append((bone.boneId, bone.locationZ(), enabled, inaccessible))

        self.boneId = bone.boneId
        # Specific debugging `if` statement
        # if Path.Log.getLevel(LOG_MODULE) == Path.Log.Level.DEBUG and bone.boneId > 2:
        #    commands = self.boneCommands(bone, False)
        # else:
        #    commands = self.boneCommands(bone, enabled)
        commands = self.boneCommands(bone, enabled)
        bone.commands = commands

        self.shapes[bone.boneId] = self.boneShapes
        Path.Log.debug(
            "<----------------------------------- %d --------------------------------------"
            % bone.boneId
        )
        return commands

    def removePathCrossing(self, commands, bone1, bone2):
        commands.append(bone2.lastCommand)
        bones = bone2.commands
        if True and hasattr(bone1, "outCommands") and hasattr(bone2, "inCommands"):
            inEdges = edgesForCommands(bone1.outCommands, bone1.tip)
            outEdges = edgesForCommands(bone2.inCommands, bone2.inChord.Start)
            for i in range(len(inEdges)):
                e1 = inEdges[i]
                for j in range(len(outEdges) - 1, -1, -1):
                    e2 = outEdges[j]
                    cutoff = DraftGeomUtils.findIntersection(e1, e2)
                    for pt in cutoff:
                        # debugCircle(e1.Curve.Center, e1.Curve.Radius, "bone.%d-1" % (self.boneId), (1.,0.,0.))
                        # debugCircle(e2.Curve.Center, e2.Curve.Radius, "bone.%d-2" % (self.boneId), (0.,1.,0.))
                        if Path.Geom.pointsCoincide(
                            pt, e1.valueAt(e1.LastParameter)
                        ) or Path.Geom.pointsCoincide(pt, e2.valueAt(e2.FirstParameter)):
                            continue
                        # debugMarker(pt, "it", (0.0, 1.0, 1.0))
                        # 1. remove all redundant commands
                        commands = commands[: -(len(inEdges) - i)]
                        # 2., correct where c1 ends
                        c1 = bone1.outCommands[i]
                        c1Params = c1.Parameters
                        c1Params.update({"X": pt.x, "Y": pt.y, "Z": pt.z})
                        c1 = Path.Command(c1.Name, c1Params)
                        commands.append(c1)
                        # 3. change where c2 starts, this depends on the command itself
                        c2 = bone2.inCommands[j]
                        if c2.Name in Path.Geom.CmdMoveArc:
                            center = e2.Curve.Center
                            offset = center - pt
                            c2Params = c2.Parameters
                            c2Params.update({"I": offset.x, "J": offset.y, "K": offset.z})
                            c2 = Path.Command(c2.Name, c2Params)
                            bones = [c2]
                            bones.extend(bone2.commands[j + 1 :])
                        else:
                            bones = bone2.commands[j:]
                        # there can only be the one ...
                        return commands, bones

        return commands, bones

    def execute(self, obj, forReal=True):
        if not obj.Base:
            return
        if forReal and not obj.Base.isDerivedFrom("Path::Feature"):
            return
        if not obj.Base.Path:
            return
        if not obj.Base.Path.Commands:
            return

        self.setup(obj, False)

        commands = []  # the dressed commands
        lastChord = Chord()  # the last chord
        lastCommand = None  # the command that generated the last chord
        lastBone = None  # track last bone for optimizations
        oddsAndEnds = []  # track chords that are connected to plunges - in case they form a loop

        boneId = 1
        self.bones = []
        self.locationBlacklist = set()
        self.length = 0
        # boneIserted = False

        for i, thisCommand in enumerate(PathUtils.getPathWithPlacement(obj.Base).Commands):
            # if i > 14:
            #    if lastCommand:
            #        commands.append(lastCommand)
            #        lastCommand = None
            #    commands.append(thisCommand)
            #    continue
            Path.Log.info("%3d: %s" % (i, thisCommand))
            if thisCommand.Name in movecommands:
                thisChord = lastChord.moveToParameters(thisCommand.Parameters)
                thisIsACandidate = self.canAttachDogbone(thisCommand, thisChord)

                if (
                    thisIsACandidate
                    and lastCommand
                    and self.shouldInsertDogbone(obj, lastChord, thisChord)
                ):
                    Path.Log.info("  Found bone corner: {}".format(lastChord.End))
                    bone = Bone(
                        boneId,
                        obj,
                        lastCommand,
                        lastChord,
                        thisChord,
                        Smooth.InAndOut,
                        thisCommand.Parameters.get("F"),
                    )
                    bones = self.insertBone(bone)
                    boneId += 1
                    if lastBone:
                        Path.Log.info("  removing potential path crossing")
                        # debugMarker(thisChord.Start, "it", (1.0, 0.0, 1.0))
                        commands, bones = self.removePathCrossing(commands, lastBone, bone)
                    commands.extend(bones[:-1])
                    lastCommand = bones[-1]
                    lastBone = bone
                elif lastCommand and thisChord.isAPlungeMove():
                    Path.Log.info("  Looking for connection in odds and ends")
                    haveNewLastCommand = False
                    for chord in (chord for chord in oddsAndEnds if lastChord.connectsTo(chord)):
                        if self.shouldInsertDogbone(obj, lastChord, chord):
                            Path.Log.info("    and there is one")
                            Path.Log.debug("    odd/end={} last={}".format(chord, lastChord))
                            bone = Bone(
                                boneId,
                                obj,
                                lastCommand,
                                lastChord,
                                chord,
                                Smooth.In,
                                lastCommand.Parameters.get("F"),
                            )
                            bones = self.insertBone(bone)
                            boneId += 1
                            if lastBone:
                                Path.Log.info("    removing potential path crossing")
                                # debugMarker(chord.Start, "it", (0.0, 1.0, 1.0))
                                commands, bones = self.removePathCrossing(commands, lastBone, bone)
                            commands.extend(bones[:-1])
                            lastCommand = bones[-1]
                            haveNewLastCommand = True
                    if not haveNewLastCommand:
                        commands.append(lastCommand)
                    lastCommand = None
                    commands.append(thisCommand)
                    lastBone = None
                elif thisIsACandidate:
                    Path.Log.info("  is a candidate, keeping for later")
                    if lastCommand:
                        commands.append(lastCommand)
                    lastCommand = thisCommand
                    lastBone = None
                elif thisChord.isANoopMove():
                    Path.Log.info("  ignoring and dropping noop move")
                    continue
                else:
                    Path.Log.info("  nope")
                    if lastCommand:
                        commands.append(lastCommand)
                        lastCommand = None
                    commands.append(thisCommand)
                    lastBone = None

                if lastChord.isAPlungeMove() and thisIsACandidate:
                    Path.Log.info("  adding to odds and ends")
                    oddsAndEnds.append(thisChord)

                lastChord = thisChord
            else:
                if thisCommand.Name[0] != "(":
                    Path.Log.info("  Clean slate")
                    if lastCommand:
                        commands.append(lastCommand)
                        lastCommand = None
                    lastBone = None
                commands.append(thisCommand)
        # for cmd in commands:
        #    Path.Log.debug("cmd = '%s'" % cmd)
        path = Path.Path(commands)
        obj.Path = path

    def setup(self, obj, initial):
        Path.Log.info("Here we go ... ")
        if initial:
            if hasattr(obj.Base, "BoneBlacklist"):
                # dressing up a bone dressup
                obj.Side = obj.Base.Side
            else:
                Path.Log.info("Default side = right")
                # otherwise dogbones are opposite of the base path's side
                side = Side.Right
                if hasattr(obj.Base, "Side") and obj.Base.Side == "Inside":
                    Path.Log.info("inside -> side = left")
                    side = Side.Left
                else:
                    Path.Log.info("not inside -> side stays right")
                if hasattr(obj.Base, "Direction") and obj.Base.Direction == "CCW":
                    Path.Log.info("CCW -> switch sides")
                    side = Side.oppositeOf(side)
                else:
                    Path.Log.info("CW -> stay on side")
                obj.Side = side

        self.toolRadius = 5
        tc = PathDressup.toolController(obj.Base)
        if tc is None or tc.ToolNumber == 0:
            self.toolRadius = 5
        else:
            tool = tc.Proxy.getTool(tc)  # PathUtils.getTool(obj, tc.ToolNumber)
            if not tool or float(tool.Diameter) == 0:
                self.toolRadius = 5
            else:
                self.toolRadius = float(tool.Diameter) / 2

        self.shapes = {}
        self.dbg = []

    def boneStateList(self, obj):
        state = {}
        # If the receiver was loaded from file, then it never generated the bone list.
        if not hasattr(self, "bones"):
            self.execute(obj)
        for nr, loc, enabled, inaccessible in self.bones:
            item = state.get((loc[0], loc[1]))
            if item:
                item[2].append(nr)
                item[3].append(loc[2])
            else:
                state[(loc[0], loc[1])] = (enabled, inaccessible, [nr], [loc[2]])
        return state


class Marker(object):
    def __init__(self, pt, r, h):
        if Path.Geom.isRoughly(h, 0):
            h = 0.1
        self.pt = pt
        self.r = r
        self.h = h
        self.sep = coin.SoSeparator()
        self.pos = coin.SoTranslation()
        self.pos.translation = (pt.x, pt.y, pt.z + h / 2)
        self.rot = coin.SoRotationXYZ()
        self.rot.axis = self.rot.X
        self.rot.angle = math.pi / 2
        self.cyl = coin.SoCylinder()
        self.cyl.radius = r
        self.cyl.height = h
        # self.cyl.removePart(self.cyl.TOP)
        # self.cyl.removePart(self.cyl.BOTTOM)
        self.material = coin.SoMaterial()
        self.sep.addChild(self.pos)
        self.sep.addChild(self.rot)
        self.sep.addChild(self.material)
        self.sep.addChild(self.cyl)
        self.lowlight()

    def setSelected(self, selected):
        if selected:
            self.highlight()
        else:
            self.lowlight()

    def highlight(self):
        self.material.diffuseColor = self.color(1)
        self.material.transparency = 0.45

    def lowlight(self):
        self.material.diffuseColor = self.color(0)
        self.material.transparency = 0.75

    def color(self, id):
        if id == 1:
            return coin.SbColor(0.9, 0.9, 0.5)
        return coin.SbColor(0.9, 0.5, 0.9)


class TaskPanel(object):
    DataIds = QtCore.Qt.ItemDataRole.UserRole
    DataKey = QtCore.Qt.ItemDataRole.UserRole + 1
    DataLoc = QtCore.Qt.ItemDataRole.UserRole + 2

    def __init__(self, viewProvider, obj):
        self.viewProvider = viewProvider
        self.obj = obj
        self.form = FreeCADGui.PySideUic.loadUi(":/panels/DogboneEdit.ui")
        self.s = None
        FreeCAD.ActiveDocument.openTransaction("Edit Dogbone Dress-up")
        self.height = 10
        self.markers = []

    def reject(self):
        FreeCAD.ActiveDocument.abortTransaction()
        FreeCADGui.Control.closeDialog()
        FreeCAD.ActiveDocument.recompute()
        FreeCADGui.Selection.removeObserver(self.s)
        self.cleanup()

    def accept(self):
        self.getFields()
        FreeCAD.ActiveDocument.commitTransaction()
        FreeCADGui.ActiveDocument.resetEdit()
        FreeCADGui.Control.closeDialog()
        FreeCAD.ActiveDocument.recompute()
        FreeCADGui.Selection.removeObserver(self.s)
        FreeCAD.ActiveDocument.recompute()
        self.cleanup()

    def cleanup(self):
        self.viewProvider.showMarkers(False)
        for m in self.markers:
            self.viewProvider.switch.removeChild(m.sep)
        self.markers = []

    def getFields(self):
        self.obj.Style = str(self.form.styleCombo.currentText())
        self.obj.Side = str(self.form.sideCombo.currentText())
        self.obj.Incision = str(self.form.incisionCombo.currentText())
        self.obj.Custom = self.form.custom.value()
        blacklist = []
        for i in range(0, self.form.bones.count()):
            item = self.form.bones.item(i)
            if item.checkState() == QtCore.Qt.CheckState.Unchecked:
                blacklist.extend(item.data(self.DataIds))
        self.obj.BoneBlacklist = sorted(blacklist)
        self.obj.Proxy.execute(self.obj)

    def updateBoneList(self):
        itemList = []
        for loc, (enabled, inaccessible, ids, zs) in self.obj.Proxy.boneStateList(self.obj).items():
            lbl = "(%.2f, %.2f): %s" % (loc[0], loc[1], ",".join(str(id) for id in ids))
            item = QtGui.QListWidgetItem(lbl)
            if enabled:
                item.setCheckState(QtCore.Qt.CheckState.Checked)
            else:
                item.setCheckState(QtCore.Qt.CheckState.Unchecked)
            flags = QtCore.Qt.ItemFlag.ItemIsSelectable
            if not inaccessible:
                flags |= QtCore.Qt.ItemFlag.ItemIsEnabled | QtCore.Qt.ItemFlag.ItemIsUserCheckable
            item.setFlags(flags)
            item.setData(self.DataIds, ids)
            item.setData(self.DataKey, ids[0])
            item.setData(self.DataLoc, loc)
            itemList.append(item)
        self.form.bones.clear()
        markers = []
        for item in sorted(itemList, key=lambda item: item.data(self.DataKey)):
            self.form.bones.addItem(item)
            loc = item.data(self.DataLoc)
            r = max(self.obj.Proxy.length, 1)
            markers.append(
                Marker(
                    FreeCAD.Vector(loc[0], loc[1], min(zs)),
                    r,
                    max(1, max(zs) - min(zs)),
                )
            )
        for m in self.markers:
            self.viewProvider.switch.removeChild(m.sep)
        for m in markers:
            self.viewProvider.switch.addChild(m.sep)
        self.markers = markers

    def updateUI(self):
        customSelected = self.obj.Incision == Incision.Custom
        self.form.custom.setEnabled(customSelected)
        self.form.customLabel.setEnabled(customSelected)
        self.updateBoneList()

        if Path.Log.getLevel(Path.Log.thisModule()) == Path.Log.Level.DEBUG:
            for obj in FreeCAD.ActiveDocument.Objects:
                if obj.Name.startswith("Shape"):
                    FreeCAD.ActiveDocument.removeObject(obj.Name)
            Path.Log.info("object name %s" % self.obj.Name)
            if hasattr(self.obj.Proxy, "shapes"):
                Path.Log.info("showing shapes attribute")
                for shapes in self.obj.Proxy.shapes.values():
                    for shape in shapes:
                        Part.show(shape)
            else:
                Path.Log.info("no shapes attribute found")

    def updateModel(self):
        self.getFields()
        self.updateUI()
        FreeCAD.ActiveDocument.recompute()

    def setupCombo(self, combo, text, items):
        if items and len(items) > 0:
            for i in range(combo.count(), -1, -1):
                combo.removeItem(i)
            combo.addItems(items)
        index = combo.findText(text, QtCore.Qt.MatchFixedString)
        if index >= 0:
            combo.setCurrentIndex(index)

    def setFields(self):
        self.setupCombo(self.form.styleCombo, self.obj.Style, Style.All)
        self.setupCombo(self.form.sideCombo, self.obj.Side, Side.All)
        self.setupCombo(self.form.incisionCombo, self.obj.Incision, Incision.All)
        self.form.custom.setMinimum(0.0)
        self.form.custom.setDecimals(3)
        self.form.custom.setValue(self.obj.Custom)
        self.updateUI()

    def open(self):
        self.s = SelObserver()
        # install the function mode resident
        FreeCADGui.Selection.addObserver(self.s)

    def setupUi(self):
        self.setFields()
        # now that the form is filled, setup the signal handlers
        self.form.styleCombo.currentIndexChanged.connect(self.updateModel)
        self.form.sideCombo.currentIndexChanged.connect(self.updateModel)
        self.form.incisionCombo.currentIndexChanged.connect(self.updateModel)
        self.form.custom.valueChanged.connect(self.updateModel)
        self.form.bones.itemChanged.connect(self.updateModel)
        self.form.bones.itemSelectionChanged.connect(self.updateMarkers)

        self.viewProvider.showMarkers(True)

    def updateMarkers(self):
        index = self.form.bones.currentRow()
        for i, m in enumerate(self.markers):
            m.setSelected(i == index)


class SelObserver(object):
    def __init__(self):
        import Path.Op.Gui.Selection as PST

        PST.eselect()

    def __del__(self):
        import Path.Op.Gui.Selection as PST

        PST.clear()

    def addSelection(self, doc, obj, sub, pnt):
        FreeCADGui.doCommand("Gui.Selection.addSelection(FreeCAD.ActiveDocument." + obj + ")")
        FreeCADGui.updateGui()


class ViewProviderDressup(object):
    def __init__(self, vobj):
        self.vobj = vobj
        self.obj = None

    def attach(self, vobj):
        self.obj = vobj.Object
        if self.obj and self.obj.Base:
            for i in self.obj.Base.InList:
                if hasattr(i, "Group"):
                    group = i.Group
                    for g in group:
                        if g.Name == self.obj.Base.Name:
                            group.remove(g)
                    i.Group = group
            # FreeCADGui.ActiveDocument.getObject(obj.Base.Name).Visibility = False
        self.switch = coin.SoSwitch()
        vobj.RootNode.addChild(self.switch)

    def showMarkers(self, on):
        sw = coin.SO_SWITCH_ALL if on else coin.SO_SWITCH_NONE
        self.switch.whichChild = sw

    def claimChildren(self):
        return [self.obj.Base]

    def setEdit(self, vobj, mode=0):
        FreeCADGui.Control.closeDialog()
        panel = TaskPanel(self, vobj.Object)
        FreeCADGui.Control.showDialog(panel)
        panel.setupUi()
        return True

    def dumps(self):
        return None

    def loads(self, state):
        return None

    def onDelete(self, arg1=None, arg2=None):
        """this makes sure that the base operation is added back to the project and visible"""
        if arg1.Object and arg1.Object.Base:
            FreeCADGui.ActiveDocument.getObject(arg1.Object.Base.Name).Visibility = True
            job = PathUtils.findParentJob(arg1.Object)
            if job:
                job.Proxy.addOperation(arg1.Object.Base, arg1.Object)
            arg1.Object.Base = None
        return True


def Create(base, name="DogboneDressup"):
    """
    Create(obj, name='DogboneDressup') ... dresses the given Path.Op.Profile/PathContour object with dogbones.
    """
    obj = FreeCAD.ActiveDocument.addObject("Path::FeaturePython", name)
    dbo = ObjectDressup(obj, base)
    job = PathUtils.findParentJob(base)
    job.Proxy.addOperation(obj, base)

    if FreeCAD.GuiUp:
        obj.ViewObject.Proxy = ViewProviderDressup(obj.ViewObject)
        obj.Base.ViewObject.Visibility = False

    dbo.setup(obj, True)
    return obj


class CommandDressupDogbone(object):
    def GetResources(self):
        return {
            "Pixmap": "CAM_Dressup",
            "MenuText": QT_TRANSLATE_NOOP("CAM_DressupDogbone", "Dogbone"),
            "ToolTip": QT_TRANSLATE_NOOP(
                "CAM_DressupDogbone",
                "Creates a dogbone dress-up object from a selected toolpath",
            ),
        }

    def IsActive(self):
        if FreeCAD.ActiveDocument is not None:
            for o in FreeCAD.ActiveDocument.Objects:
                if o.Name[:3] == "Job":
                    return True
        return False

    def Activated(self):

        # check that the selection contains exactly what we want
        selection = FreeCADGui.Selection.getSelection()
        if len(selection) != 1:
            FreeCAD.Console.PrintError(
                translate("CAM_DressupDogbone", "Select one toolpath object") + "\n"
            )
            return
        baseObject = selection[0]
        if not baseObject.isDerivedFrom("Path::Feature"):
            FreeCAD.Console.PrintError(
                translate("CAM_DressupDogbone", "The selected object is not a toolpath") + "\n"
            )
            return

        # everything ok!
        FreeCAD.ActiveDocument.openTransaction("Create Dogbone Dress-up")
        FreeCADGui.addModule("Path.Dressup.Gui.Dogbone")
        FreeCADGui.doCommand(
            "Path.Dressup.Gui.Dogbone.Create(FreeCAD.ActiveDocument.%s)" % baseObject.Name
        )
        # FreeCAD.ActiveDocument.commitTransaction()  # Final `commitTransaction()` called via TaskPanel.accept()
        FreeCAD.ActiveDocument.recompute()


# obsolete, replaced by DogboneII
# if FreeCAD.GuiUp:
#    import FreeCADGui
#    from PySide import QtGui
#    from pivy import coin
#
#    FreeCADGui.addCommand("CAM_DressupDogbone", CommandDressupDogbone())

FreeCAD.Console.PrintLog("Loading DressupDogbone… done\n")



# SPDX-License-Identifier: LGPL-2.1-or-later
# ***************************************************************************
# *   Copyright (c) 2021 sliptonic <shopinthewoods@gmail.com>               *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************


# Technical Debt.  This generator currently assumes 3+2 axis rotation of CA.
# The main generator function should be extended to include other flavors of 3+2


import math
import Path
import FreeCAD
from enum import Enum

__title__ = "Rotation toolpath Generator"
__author__ = "sliptonic (Brad Collette)"
__url__ = "https://www.freecad.org"
__doc__ = "Generates the rotation toolpath"


if False:
    Path.Log.setLevel(Path.Log.Level.DEBUG, Path.Log.thisModule())
    Path.Log.trackModule(Path.Log.thisModule())
else:
    Path.Log.setLevel(Path.Log.Level.INFO, Path.Log.thisModule())


class refAxis(Enum):
    x = FreeCAD.Vector(1, 0, 0)
    y = FreeCAD.Vector(0, 1, 0)
    z = FreeCAD.Vector(0, 0, 1)


def relAngle(vec, ref):
    """
    Takes a vector and a reference axis (refAxis) vector. Calculates the
    relative angle.  The result is returned in degrees (plus or minus)
    """

    Path.Log.debug("vec: {}  ref: {}".format(vec, ref))
    norm = vec * 1  # copy vec so we don't alter original

    if ref == refAxis.x:
        plane = refAxis.y.value
    elif ref == refAxis.y:
        plane = refAxis.z.value
    else:
        plane = refAxis.x.value

    norm.projectToPlane(FreeCAD.Vector(0, 0, 0), plane)

    ref = ref.value
    rot = FreeCAD.Rotation(norm, ref)
    ang = math.degrees(rot.Angle)
    angle = ang * plane.dot(rot.Axis)
    Path.Log.debug("relative ang: {}".format(angle))

    return angle


def __getCRotation(normalVector, cMin=-360, cMax=360):
    """
    Calculate the valid C axis rotations component to align the normalVector
    with either the +y or -y axis.
    multiple poses may be possible.  Returns a list of all valid poses
    """
    Path.Log.debug("normalVector: {} cMin: {} cMax: {}".format(normalVector, cMin, cMax))

    angle = relAngle(normalVector, refAxis.y)

    # Given an angle, there are four possibilities; rotating +- to each of the
    # two axes +y and -y
    candidates = [angle]
    if angle == 0:
        candidates.append(180)
    elif angle == 180:
        candidates.append(0)
    elif angle >= 0:
        candidates.append(angle - 180)
        candidates.append(180 + angle)
        candidates.append(angle - 360)
    else:
        candidates.append(angle + 180)
        candidates.append(-180 + angle)
        candidates.append(angle + 360)

    # final results are candidates that don't violate rotation limits
    results = [c for c in candidates if c >= cMin and c <= cMax]

    return results


def __getARotation(normalVector, aMin=-360, aMax=360):
    """
    Calculate the A axis rotation component.
    Final rotation is always assumed to be around +X. The sign of the returned
    value indicates direction of rotation.

    Returns None if rotation violates min/max constraints
    """

    angle = relAngle(normalVector, refAxis.z)

    # only return a result if it doesn't violate rotation constraints
    if angle > aMin and angle <= aMax:
        return angle
    else:
        return None


def generate(normalVector, aMin=-360, aMax=360, cMin=-360, cMax=360, compound=False):
    """
    Generates Gcode rotation to align a vector (alignVector) with the positive Z axis.

    It first rotates around the Z axis (C rotation)
    to align the vector the positive Y axis. Then around the X axis
    (A rotation).

    The min and max arguments dictate the range of motion allowed rotation in
    the respective axis.
    Default assumes continuous rotation.

    Returns a list of path commands for the shortest valid solution

    If compound is False, axis moves will be broken out to individual commands

    The normalVector input from a typical face (f) can be obtained like this:

        u, v = f.ParameterRange[:2]
        n = f.normalAt(u,v)
        plm = obj.getGlobalPlacement()
        rot = plm.Rotation
        normalVector = rot.multVec(n
    """

    Path.Log.track(
        "\n=============\n normalVector: {}\n aMin: {}\n aMax: {}\n cMin: {}\n cMax: {}".format(
            normalVector, aMin, aMax, cMin, cMax
        )
    )

    # Calculate C rotation
    cResults = __getCRotation(normalVector, cMin, cMax)
    Path.Log.debug("C Rotation results {}".format(cResults))

    solutions = []
    for result in cResults:

        # calculate a new vector based on the result
        rot = FreeCAD.Rotation(FreeCAD.Vector(0, 0, 1), result)
        newvec = rot.multVec(normalVector)

        # Get the candidate A rotation for the new vector
        aResult = __getARotation(newvec, aMin, aMax)

        Path.Log.debug("\n=====\nFor C Rotation: {}\n Calculated A {}\n".format(result, aResult))

        if aResult is not None:
            solutions.append({"A": aResult, "C": result})

    if len(solutions) == 0:  # No valid solution found
        raise ValueError("No valid rotation solution found")

    # find pose with the shortest transit length
    best = solutions[0]
    curlen = math.fabs(best["A"]) + math.fabs(best["C"])
    for solution in solutions[1:]:
        testlen = math.fabs(solution["A"]) + math.fabs(solution["C"])
        if testlen < curlen:
            best = solution
            curlen = testlen

    Path.Log.debug("best result: {}".format(best))

    # format and return rotation commands
    commands = []
    if compound:
        commands.append(Path.Command("G0", best))
    else:
        for key, val in best.items():
            print(key, val)
            commands.append(Path.Command("G0", {key: val}))

    return commands



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   (c) 2009, 2010 Yorik van Havre <yorik@uncreated.net>                  *
# *   (c) 2009, 2010 Ken Cline <cline@frii.com>                             *
# *   (c) 2020 Eliud Cabrera Castillo <e.cabrera-castillo@tum.de>           *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides GUI tools to join lines and wires."""
## @package gui_join
# \ingroup draftguitools
# \brief Provides GUI tools to join lines and wires.

## \addtogroup draftguitools
# @{
from PySide.QtCore import QT_TRANSLATE_NOOP

import FreeCADGui as Gui
import Draft_rc
from draftguitools import gui_base_original
from draftguitools import gui_tool_utils
from draftutils import utils
from draftutils.messages import _msg, _err, _toolmsg
from draftutils.translate import translate

# The module is used to prevent complaints from code checkers (flake8)
True if Draft_rc.__name__ else False


class Join(gui_base_original.Modifier):
    """Gui Command for the Join tool."""

    def GetResources(self):
        """Set icon, menu and tooltip."""

        return {
            "Pixmap": "Draft_Join",
            "Accel": "J, O",
            "MenuText": QT_TRANSLATE_NOOP("Draft_Join", "Join"),
            "ToolTip": QT_TRANSLATE_NOOP(
                "Draft_Join",
                "Joins the selected lines or polylines into a single object.\nThe lines must share a common point at the start or at the end.",
            ),
        }

    def Activated(self):
        """Execute when the command is called."""
        super().Activated(name="Join")
        if not self.ui:
            return
        if not Gui.Selection.getSelection():
            self.ui.selectUi(on_close_call=self.finish)
            _msg(translate("draft", "Select an object to join"))
            self.call = self.view.addEventCallback("SoEvent", gui_tool_utils.selectObject)
        else:
            self.proceed()

    def proceed(self):
        """Proceed with execution of the command after proper selection.

        BUG: It occasionally fails to join lines even if the lines
        visually share a point. This is due to the underlying `joinWires`
        method not handling the points correctly.
        """
        if self.call is not None:
            self.end_callbacks(self.call)
        if Gui.Selection.getSelection():
            self.print_selection()
            if all(utils.get_type(o) == "Wire" for o in Gui.Selection.getSelection()):
                Gui.addModule("Draft")
                _cmd = "Draft.join_wires"
                _cmd += "("
                _cmd += "FreeCADGui.Selection.getSelection()"
                _cmd += ")"
                _cmd_list = ["j = " + _cmd, "FreeCAD.ActiveDocument.recompute()"]
                self.commit(translate("draft", "Join Lines"), _cmd_list)
            else:
                _err(translate("draft", "Only Draft lines and wires can be joined"))
        self.finish()

    def print_selection(self):
        """Print the selected items."""
        labels = []
        for obj in Gui.Selection.getSelection():
            labels.append(obj.Label)

        labels = ", ".join(labels)
        _toolmsg(translate("draft", "Selection:") + " {}".format(labels))


Gui.addCommand("Draft_Join", Join())

## @}



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2020 sliptonic <shopinthewoods@gmail.com>               *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

import FreeCAD
import Part
import Path
import CAMTests.PathTestUtils as PathTestUtils

vd = None


def initVD():
    global vd
    if vd is None:
        pts = [
            (0, 0),
            (3.5, 0),
            (3.5, 1),
            (1, 1),
            (1, 2),
            (2.5, 2),
            (2.5, 3),
            (1, 3),
            (1, 4),
            (3.5, 4),
            (3.5, 5),
            (0, 5),
        ]
        ptv = [FreeCAD.Vector(p[0], p[1]) for p in pts]
        ptv.append(ptv[0])

        vd = Path.Voronoi.Diagram()
        for i in range(len(pts)):
            vd.addSegment(ptv[i], ptv[i + 1])

        vd.construct()

        for e in vd.Edges:
            e.Color = 0 if e.isPrimary() else 1

        vd.colorExterior(2)
        vd.colorColinear(3)
        vd.colorTwins(4)


class TestPathVoronoi(PathTestUtils.PathTestBase):
    def setUp(self):
        initVD()

    def test00(self):
        """Check vertex comparison"""

        self.assertTrue(vd.Vertices[0] == vd.Vertices[0])
        self.assertTrue(vd.Vertices[1] == vd.Vertices[1])
        self.assertTrue(vd.Vertices[0] != vd.Vertices[1])
        self.assertEqual(vd.Vertices[0], vd.Vertices[0])
        self.assertEqual(vd.Vertices[1], vd.Vertices[1])
        self.assertNotEqual(vd.Vertices[0], vd.Vertices[1])
        self.assertNotEqual(vd.Vertices[1], vd.Vertices[0])

    def test10(self):
        """Check edge comparison"""

        self.assertTrue(vd.Edges[0] == vd.Edges[0])
        self.assertTrue(vd.Edges[1] == vd.Edges[1])
        self.assertTrue(vd.Edges[0] != vd.Edges[1])
        self.assertEqual(vd.Edges[0], vd.Edges[0])
        self.assertEqual(vd.Edges[1], vd.Edges[1])
        self.assertNotEqual(vd.Edges[0], vd.Edges[1])
        self.assertNotEqual(vd.Edges[1], vd.Edges[0])

    def test20(self):
        """Check cell comparison"""

        self.assertTrue(vd.Cells[0] == vd.Cells[0])
        self.assertTrue(vd.Cells[1] == vd.Cells[1])
        self.assertTrue(vd.Cells[0] != vd.Cells[1])
        self.assertEqual(vd.Cells[0], vd.Cells[0])
        self.assertEqual(vd.Cells[1], vd.Cells[1])
        self.assertNotEqual(vd.Cells[0], vd.Cells[1])
        self.assertNotEqual(vd.Cells[1], vd.Cells[0])

    def test50(self):
        """Check toShape for linear edges"""

        edges = [e for e in vd.Edges if e.Color == 0 and e.isLinear()]
        self.assertNotEqual(len(edges), 0)
        e0 = edges[0]

        e = e0.toShape()
        self.assertTrue(type(e.Curve) == Part.LineSegment or type(e.Curve) == Part.Line)
        self.assertFalse(
            Path.Geom.pointsCoincide(e.valueAt(e.FirstParameter), e.valueAt(e.LastParameter))
        )
        self.assertRoughly(e.valueAt(e.FirstParameter).z, 0)
        self.assertRoughly(e.valueAt(e.LastParameter).z, 0)

    def test51(self):
        """Check toShape for linear edges with set z"""

        edges = [e for e in vd.Edges if e.Color == 0 and e.isLinear()]
        self.assertNotEqual(len(edges), 0)
        e0 = edges[0]

        e = e0.toShape(13.7)
        self.assertTrue(type(e.Curve) == Part.LineSegment or type(e.Curve) == Part.Line)
        self.assertFalse(
            Path.Geom.pointsCoincide(e.valueAt(e.FirstParameter), e.valueAt(e.LastParameter))
        )
        self.assertRoughly(e.valueAt(e.FirstParameter).z, 13.7)
        self.assertRoughly(e.valueAt(e.LastParameter).z, 13.7)

    def test52(self):
        """Check toShape for linear edges with varying z"""

        edges = [e for e in vd.Edges if e.Color == 0 and e.isLinear()]
        self.assertNotEqual(len(edges), 0)
        e0 = edges[0]

        e = e0.toShape(2.37, 5.14)
        self.assertTrue(type(e.Curve) == Part.LineSegment or type(e.Curve) == Part.Line)
        self.assertFalse(
            Path.Geom.pointsCoincide(e.valueAt(e.FirstParameter), e.valueAt(e.LastParameter))
        )
        self.assertRoughly(e.valueAt(e.FirstParameter).z, 2.37)
        self.assertRoughly(e.valueAt(e.LastParameter).z, 5.14)

    def test60(self):
        """Check toShape for curved edges"""

        edges = [e for e in vd.Edges if e.Color == 0 and e.isCurved()]
        self.assertNotEqual(len(edges), 0)
        e0 = edges[0]

        e = e0.toShape()
        self.assertTrue(type(e.Curve) == Part.Parabola or type(e.Curve) == Part.BSplineCurve)
        self.assertFalse(
            Path.Geom.pointsCoincide(e.valueAt(e.FirstParameter), e.valueAt(e.LastParameter))
        )
        self.assertRoughly(e.valueAt(e.FirstParameter).z, 0)
        self.assertRoughly(e.valueAt(e.LastParameter).z, 0)

    def test61(self):
        """Check toShape for curved edges with set z"""

        edges = [e for e in vd.Edges if e.Color == 0 and e.isCurved()]
        self.assertNotEqual(len(edges), 0)
        e0 = edges[0]

        e = e0.toShape(13.7)
        self.assertTrue(type(e.Curve) == Part.Parabola or type(e.Curve) == Part.BSplineCurve)
        self.assertFalse(
            Path.Geom.pointsCoincide(e.valueAt(e.FirstParameter), e.valueAt(e.LastParameter))
        )
        self.assertRoughly(e.valueAt(e.FirstParameter).z, 13.7)
        self.assertRoughly(e.valueAt(e.LastParameter).z, 13.7)

    def test62(self):
        """Check toShape for curved edges with varying z"""

        edges = [e for e in vd.Edges if e.Color == 0 and e.isCurved()]
        self.assertNotEqual(len(edges), 0)
        e0 = edges[0]

        e = e0.toShape(2.37, 5.14)
        self.assertTrue(type(e.Curve) == Part.Parabola or type(e.Curve) == Part.BSplineCurve)
        self.assertFalse(
            Path.Geom.pointsCoincide(e.valueAt(e.FirstParameter), e.valueAt(e.LastParameter))
        )
        self.assertRoughly(e.valueAt(e.FirstParameter).z, 2.37)
        self.assertRoughly(e.valueAt(e.LastParameter).z, 5.14)



# ***************************************************************************
# *   Copyright (c) 2020 Sudhanshu Dubey <sudhanshu.thethunder@gmail.com>   *
# *   Copyright (c) 2020 Bernd Hahnebach <bernd@bimstatik.org>              *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   This program is distributed in the hope that it will be useful,       *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with this program; if not, write to the Free Software   *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************

import math

import FreeCAD
from FreeCAD import Rotation
from FreeCAD import Vector

import Part
import Sketcher
from BOPTools.SplitFeatures import makeBooleanFragments
from BOPTools.SplitFeatures import makeSlice
from CompoundTools.CompoundFilter import makeCompoundFilter

import Fem
import ObjectsFem

from . import manager
from .manager import get_meshname
from .manager import init_doc
from .meshes import generate_mesh


def get_information():
    return {
        "name": "Constraint Section Print",
        "meshtype": "solid",
        "meshelement": "Tet10",
        "constraints": ["section_print", "fixed", "pressure"],
        "solvers": ["ccxtools"],
        "material": "solid",
        "equations": ["mechanical"],
    }


def get_explanation(header=""):
    return (
        header
        + """

To run the example from Python console use:
from femexamples.constraint_section_print import setup
setup()


See forum topic post:
https://forum.freecad.org/viewtopic.php?t=43044

constraint section print with volume elements

"""
    )


def setup(doc=None, solvertype="ccxtools"):

    # init FreeCAD document
    if doc is None:
        doc = init_doc()

    # explanation object
    # just keep the following line and change text string in get_explanation method
    manager.add_explanation_obj(doc, get_explanation(manager.get_header(get_information())))

    # geometric objects
    # the part sketch
    arc_sketch = doc.addObject("Sketcher::SketchObject", "Arc_Sketch")
    arc_sketch.Placement = FreeCAD.Placement(Vector(0, 0, 0), Rotation(0, 0, 0, 1))
    arc_sketch.MapMode = "Deactivated"
    # not the exact geometry which makes a closed wire
    # exact geometry will be made by the constraints
    # the order is important for the constraints definition
    geoList = [
        Part.ArcOfCircle(Part.Circle(Vector(0, 0, 0), Vector(0, 0, 1), 47), 0, math.pi),
        Part.ArcOfCircle(
            Part.Circle(Vector(-19, -22, 0), Vector(0, 0, 1), 89),
            math.pi / 12,
            math.pi / 1.1,
        ),
        Part.LineSegment(Vector(-105, 0, 0), Vector(-47, 0, 0)),
        Part.LineSegment(Vector(47, 0, 0), Vector(67, 0, 0)),
    ]
    arc_sketch.addGeometry(geoList, False)
    # https://wiki.freecad.org/Sketcher_ConstrainCoincident
    # but the best way is to add a constraint is watching
    # FreeCAD python console while create this one by the Gui
    conList = [
        Sketcher.Constraint("Coincident", 0, 3, -1, 1),
        Sketcher.Constraint("PointOnObject", 0, 2, -1),
        Sketcher.Constraint("PointOnObject", 0, 1, -1),
        Sketcher.Constraint("PointOnObject", 1, 2, -1),
        Sketcher.Constraint("PointOnObject", 1, 1, -1),
        Sketcher.Constraint("Coincident", 2, 1, 0, 2),
        Sketcher.Constraint("Coincident", 2, 2, 1, 2),
        Sketcher.Constraint("Coincident", 3, 1, 1, 1),
        Sketcher.Constraint("Coincident", 3, 2, 0, 1),
        Sketcher.Constraint("DistanceX", 2, 2, 2, 1, 58),
        Sketcher.Constraint("DistanceX", 3, 2, 3, 1, 20),
        Sketcher.Constraint("Radius", 0, 47),
        Sketcher.Constraint("Radius", 1, 89),
    ]
    arc_sketch.addConstraint(conList)

    # the part extrusion
    extrude_part = doc.addObject("Part::Extrusion", "ArcExtrude")
    extrude_part.Base = arc_sketch
    extrude_part.DirMode = "Custom"
    extrude_part.Dir = Vector(0.00, 0.00, 1.00)
    extrude_part.DirLink = None
    extrude_part.LengthFwd = 30.00
    extrude_part.LengthRev = 0.00
    extrude_part.Solid = True
    extrude_part.Reversed = False
    extrude_part.Symmetric = True
    extrude_part.TaperAngle = 0.00
    extrude_part.TaperAngleRev = 0.00

    # section plane sketch
    section_sketch = doc.addObject("Sketcher::SketchObject", "Section_Sketch")
    section_sketch.Placement = FreeCAD.Placement(
        Vector(0.000000, 0.000000, 0.000000),
        Rotation(0.000000, 0.000000, 0.000000, 1.000000),
    )
    section_sketch.MapMode = "Deactivated"
    section_sketch.addGeometry(
        Part.LineSegment(Vector(-6.691961, -16.840161, 0), Vector(75.156087, 79.421394, 0)),
        False,
    )
    # section_sketch.ExternalGeometry = extrude_part

    # section plane extrusion
    extrude_section_plane = doc.addObject("Part::Extrusion", "SectionPlaneExtrude")
    extrude_section_plane.Base = section_sketch
    extrude_section_plane.DirMode = "Custom"
    extrude_section_plane.Dir = Vector(0.00, 0.00, -1.00)
    extrude_section_plane.DirLink = None
    extrude_section_plane.LengthFwd = 40.00
    extrude_section_plane.LengthRev = 0.00
    extrude_section_plane.Solid = False
    extrude_section_plane.Reversed = False
    extrude_section_plane.Symmetric = True
    extrude_section_plane.TaperAngle = 0.00
    extrude_section_plane.TaperAngleRev = 0.00

    # TODO the extrusions could be done with much less code, see BOLTS

    if FreeCAD.GuiUp:
        arc_sketch.ViewObject.hide()
        section_sketch.ViewObject.hide()
        extrude_part.ViewObject.hide()
        extrude_section_plane.ViewObject.hide()

    Slice = makeSlice(name="Slice")
    Slice.Base = extrude_part
    Slice.Tools = extrude_section_plane
    Slice.Mode = "Split"
    # Slice.Proxy.execute(Slice)
    Slice.purgeTouched()

    # compound filter to get the solids of the slice
    solid_one = makeCompoundFilter(name="SolidOne")
    solid_one.Base = Slice
    solid_one.FilterType = "specific items"
    solid_one.items = "0"
    # solid_one.Proxy.execute(solid_one)
    solid_one.purgeTouched()
    if FreeCAD.GuiUp:
        solid_one.Base.ViewObject.hide()

    solid_two = makeCompoundFilter(name="SolidTwo")
    solid_two.Base = Slice
    solid_two.FilterType = "specific items"
    solid_two.items = "1"
    # solid_two.Proxy.execute(solid_two)
    solid_two.purgeTouched()
    if FreeCAD.GuiUp:
        solid_two.Base.ViewObject.hide()

    # CompSolid out of the two solids
    geom_obj = makeBooleanFragments(name="BooleanFragments")
    geom_obj.Objects = [solid_one, solid_two]
    geom_obj.Mode = "CompSolid"
    # geom_obj.Proxy.execute(geom_obj)
    geom_obj.purgeTouched()
    if FreeCAD.GuiUp:
        solid_one.ViewObject.hide()
        solid_two.ViewObject.hide()
    doc.recompute()

    if FreeCAD.GuiUp:
        geom_obj.ViewObject.Transparency = 50
        geom_obj.ViewObject.Document.activeView().viewAxonometric()
        geom_obj.ViewObject.Document.activeView().fitAll()

    # analysis
    analysis = ObjectsFem.makeAnalysis(doc, "Analysis")

    # solver
    if solvertype == "ccxtools":
        solver_obj = ObjectsFem.makeSolverCalculiXCcxTools(doc, "CalculiXCcxTools")
        solver_obj.WorkingDir = ""
    else:
        FreeCAD.Console.PrintWarning(
            "Unknown or unsupported solver type: {}. "
            "No solver object was created.\n".format(solvertype)
        )
    if solvertype == "ccxtools":
        solver_obj.SplitInputWriter = False
        solver_obj.AnalysisType = "static"
        solver_obj.GeometricalNonlinearity = "linear"
        solver_obj.ThermoMechSteadyState = False
        solver_obj.MatrixSolverType = "default"
        solver_obj.IterationsControlParameterTimeUse = False
    analysis.addObject(solver_obj)

    # material
    material_obj = ObjectsFem.makeMaterialSolid(doc, "Material")
    mat = material_obj.Material
    mat["Name"] = "CalculiX-Steel"
    mat["YoungsModulus"] = "210000 MPa"
    mat["PoissonRatio"] = "0.30"
    material_obj.Material = mat
    analysis.addObject(material_obj)

    # constraint fixed
    con_fixed = ObjectsFem.makeConstraintFixed(doc, "ConstraintFixed")
    con_fixed.References = [(geom_obj, "Face9")]
    analysis.addObject(con_fixed)

    # constraint pressure
    con_pressure = ObjectsFem.makeConstraintPressure(doc, "ConstraintPressure")
    con_pressure.References = [(geom_obj, "Face1")]
    con_pressure.Pressure = "100.0 MPa"
    con_pressure.Reversed = False
    analysis.addObject(con_pressure)

    # constraint section print
    con_sectionpr = ObjectsFem.makeConstraintSectionPrint(doc, "ConstraintSectionPrint")
    con_sectionpr.References = [(geom_obj, "Face6")]
    analysis.addObject(con_sectionpr)

    # mesh
    from .meshes.mesh_section_print_tetra10 import create_nodes, create_elements

    fem_mesh = generate_mesh.mesh_from_existing(create_nodes, create_elements)
    femmesh_obj = analysis.addObject(ObjectsFem.makeMeshGmsh(doc, get_meshname()))[0]
    femmesh_obj.FemMesh = fem_mesh
    femmesh_obj.Shape = geom_obj
    femmesh_obj.SecondOrderLinear = False

    doc.recompute()
    return doc



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *   Copyright (c) 2009, 2010 Yorik van Havre <yorik@uncreated.net>        *
# *   Copyright (c) 2009, 2010 Ken Cline <cline@frii.com>                   *
# *   Copyright (c) 2020 Eliud Cabrera Castillo <e.cabrera-castillo@tum.de> *
# *   Copyright (c) 2025 FreeCAD Project Association                        *
# *                                                                         *
# *   This file is part of the FreeCAD CAx development system.              *
# *                                                                         *
# *   This program is free software; you can redistribute it and/or modify  *
# *   it under the terms of the GNU Lesser General Public License (LGPL)    *
# *   as published by the Free Software Foundation; either version 2 of     *
# *   the License, or (at your option) any later version.                   *
# *   for detail see the LICENCE text file.                                 *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful,            *
# *   but WITHOUT ANY WARRANTY; without even the implied warranty of        *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the         *
# *   GNU Library General Public License for more details.                  *
# *                                                                         *
# *   You should have received a copy of the GNU Library General Public     *
# *   License along with FreeCAD; if not, write to the Free Software        *
# *   Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  *
# *   USA                                                                   *
# *                                                                         *
# ***************************************************************************
"""Provides GUI tools to do various operations with groups.

For example, add objects to groups, select objects inside groups,
set the automatic group in which to create objects, and add objects
to the construction group.
"""
## @package gui_groups
# \ingroup draftguitools
# \brief Provides GUI tools to do various operations with groups.

## \addtogroup draftguitools
# @{
from PySide import QtCore
from PySide import QtWidgets
from PySide.QtCore import QT_TRANSLATE_NOOP

import FreeCAD as App
import FreeCADGui as Gui
import Draft_rc
from draftguitools import gui_base
from draftmake import make_layer
from draftutils import groups
from draftutils import params
from draftutils import utils
from draftutils.translate import translate


# The module is used to prevent complaints from code checkers (flake8)
True if Draft_rc.__name__ else False


class AddToGroup(gui_base.GuiCommandNeedsSelection):
    """GuiCommand for the Draft_AddToGroup tool.

    It adds selected objects to a group, or removes them from any group.

    It inherits `GuiCommandNeedsSelection` to only be available
    when there is a document and a selection.
    See this class for more information.
    """

    def __init__(self):
        super().__init__(name="Draft_AddToGroup")

    def GetResources(self):
        """Set icon, menu and tooltip."""
        return {
            "Pixmap": "Draft_AddToGroup",
            "MenuText": QT_TRANSLATE_NOOP("Draft_AddToGroup", "Add to Group"),
            "ToolTip": QT_TRANSLATE_NOOP(
                "Draft_AddToGroup",
                "Adds selected objects to a group, or removes them from any group",
            ),
        }

    def Activated(self):
        """Execute when the command is called."""
        super().Activated()

        if not hasattr(Gui, "draftToolBar"):
            return

        self.ui = Gui.draftToolBar
        objs = [obj for obj in self.doc.Objects if groups.is_group(obj)]
        objs.sort(key=lambda obj: obj.Label)
        self.objects = [None] + [None] + objs
        self.labels = (
            [translate("draft", "Ungroup")]
            + ["---"]
            + [obj.Label for obj in objs]
            + ["---"]
            + [translate("draft", "Add to New Group")]
        )
        self.icons = (
            [self.ui.getIcon(":/icons/list-remove.svg")]
            + [None]
            + [obj.ViewObject.Icon for obj in objs]
            + [None]
            + [self.ui.getIcon(":/icons/list-add.svg")]
        )

        # It uses the `DraftToolBar` class defined in the `DraftGui` module and
        # globally initialized in the `Gui` namespace to pop up a menu.
        # Once the desired option is chosen it launches the `proceed` method.
        self.ui.sourceCmd = self
        self.ui.popupMenu(self.labels, self.icons)

    def proceed(self, option):
        """Place the selected objects in the chosen group or ungroup them.
        Parameters
        ----------
        option: str
            The passed string.
        """
        self.ui.sourceCmd = None

        if option == self.labels[0]:
            # "Ungroup"
            self.doc.openTransaction(translate("draft", "Ungroup"))
            for obj in Gui.Selection.getSelection():
                try:
                    groups.ungroup(obj)
                except Exception:
                    pass
            self.doc.commitTransaction()
            self.doc.recompute()
            return

        if option == self.labels[-1]:
            # "Add to new group..."
            grp = AddNamedGroup()  # handles transaction
            grp.Activated()
            return

        # Group has been selected
        self.doc.openTransaction(translate("draft", "Add to Group"))
        i = self.labels.index(option)
        grp = self.objects[i]
        moveToGroup(grp)
        self.doc.commitTransaction()
        self.doc.recompute()


Gui.addCommand("Draft_AddToGroup", AddToGroup())


def moveToGroup(group):
    """
    Place the selected objects in the chosen group.
    """

    for obj in Gui.Selection.getSelection():
        try:
            # retrieve group's visibility
            obj.ViewObject.Visibility = group.ViewObject.Visibility
            group.addObject(obj)

        except Exception:
            pass


class SelectGroup(gui_base.GuiCommandNeedsSelection):
    """GuiCommand for the Draft_SelectGroup tool."""

    def __init__(self):
        super().__init__(name="Draft_SelectGroup")

    def GetResources(self):
        """Set icon, menu and tooltip."""
        return {
            "Pixmap": "Draft_SelectGroup",
            "MenuText": QT_TRANSLATE_NOOP("Draft_SelectGroup", "Select Group"),
            "ToolTip": QT_TRANSLATE_NOOP(
                "Draft_SelectGroup",
                "Selects the contents of selected groups. For selected non-group objects, the contents of the group they are in are selected.",
            ),
        }

    def Activated(self):
        """Execute when the command is called."""
        super().Activated()

        sel = Gui.Selection.getSelection()
        subs = []
        for obj in sel:
            if groups.is_group(obj):
                for sub in obj.Group:
                    subs.append(sub)
            else:
                for parent in obj.InList:
                    if groups.is_group(parent):
                        for sub in parent.Group:
                            subs.append(sub)

        # Always clear the selection:
        Gui.Selection.clearSelection()

        # Create a new selection from the sub objects:
        for sub in subs:
            Gui.Selection.addSelection(sub)

        # Inform the user if there is no new selection:
        if not Gui.Selection.hasSelection():
            msg = translate(
                "draft", "No new selection. Select non-empty groups or objects inside groups."
            )
            App.Console.PrintMessage(msg + "\n")


Gui.addCommand("Draft_SelectGroup", SelectGroup())


class SetAutoGroup(gui_base.GuiCommandSimplest):
    """GuiCommand for the Draft_AutoGroup tool."""

    def __init__(self):
        super().__init__(name="Draft_AutoGroup")

    def GetResources(self):
        """Set icon, menu and tooltip."""
        return {
            "Pixmap": "Draft_AutoGroup",
            "MenuText": QT_TRANSLATE_NOOP("Draft_AutoGroup", "Auto-Group"),
            "ToolTip": QT_TRANSLATE_NOOP(
                "Draft_AutoGroup", "Adds new Draft and BIM objects to the selected layer or group"
            ),
        }

    def Activated(self):
        """Execute when the command is called.

        It calls the `setAutogroup` method of the `DraftToolBar` class
        installed inside the global `Gui` namespace.
        """
        super().Activated()

        if not hasattr(Gui, "draftToolBar"):
            return

        # It uses the `DraftToolBar` class defined in the `DraftGui` module and
        # globally initialized in the `Gui` namespace to run some actions.
        # If a layer or group is selected, it runs the `AutoGroup` method.
        self.ui = Gui.draftToolBar
        sel = Gui.Selection.getSelection()
        if len(sel) == 1:
            if utils.get_type(sel[0]) == "Layer" or (
                params.get_param("AutogroupAddGroups") and groups.is_group(sel[0])
            ):
                self.ui.setAutoGroup(sel[0].Name)
                return

        # Otherwise it builds a list of layers and groups, with labels and icons,
        # including the options "None" and "Add new layer".
        if params.get_param("AutogroupAddGroups"):
            grps = [obj for obj in self.doc.Objects if groups.is_group(obj)]
            grps.sort(key=lambda obj: obj.Label)
        else:
            grps = []
        lyrs = [obj for obj in self.doc.Objects if utils.get_type(obj) == "Layer"]
        lyrs.sort(key=lambda obj: obj.Label)
        self.names = (
            [None] + [None] + [obj.Name for obj in grps] + [None] + [obj.Name for obj in lyrs]
        )
        self.labels = (
            [translate("draft", "None")]
            + ["---"]
            + [obj.Label for obj in grps]
            + ["---"]
            + [obj.Label for obj in lyrs]
            + ["---"]
            + [translate("draft", "New Layer")]
        )
        self.icons = (
            [self.ui.getIcon(":/icons/button_invalid.svg")]
            + [None]
            + [obj.ViewObject.Icon for obj in grps]
            + [None]
            + [obj.ViewObject.Icon for obj in lyrs]
            + [None]
            + [self.ui.getIcon(":/icons/Draft_NewLayer.svg")]
        )

        # With the created lists it uses the interface to pop up a menu with options.
        # Once the desired option is chosen it launches the `proceed` method.
        self.ui.sourceCmd = self
        pos = self.ui.autoGroupButton.mapToGlobal(
            QtCore.QPoint(0, self.ui.autoGroupButton.geometry().height())
        )
        self.ui.popupMenu(self.labels, self.icons, pos)

    def proceed(self, option):
        """Set the defined autogroup, or create a new layer.

        Parameters
        ----------
        option: str
            The passed string.
        """
        self.ui.sourceCmd = None

        if option == self.labels[0]:
            # "None"
            self.ui.setAutoGroup(None)
            return

        if option == self.labels[-1]:
            # "New layer..."
            txt, ok = QtWidgets.QInputDialog.getText(
                None,
                translate("draft", "New Layer"),
                translate("draft", "Layer name"),
                text=translate("draft", "Layer", "Object label"),
            )
            if not ok:
                return
            if not txt:
                return
            self.doc.openTransaction(translate("draft", "New layer"))
            lyr = make_layer.make_layer(
                name=txt,
                line_color=None,
                shape_color=None,
                line_width=None,
                draw_style=None,
                transparency=None,
            )
            self.doc.commitTransaction()
            self.doc.recompute()
            self.ui.setAutoGroup(lyr.Name)  # this...
            # self.ui.autoGroupButton.setDown(False) # or this?
            return

        # Layer or group has been selected
        i = self.labels.index(option)
        self.ui.setAutoGroup(self.names[i])


Gui.addCommand("Draft_AutoGroup", SetAutoGroup())


class AddToConstruction(gui_base.GuiCommandNeedsSelection):
    """GuiCommand for the Draft_AddConstruction tool.

    It adds the selected objects to the construction group
    defined in the `DraftToolBar` class which is initialized
    in the `Gui` namespace when the workbench loads.

    It adds a construction group if it doesn't exist.

    Added objects are also given the visual properties of the construction
    group.
    """

    def __init__(self):
        super().__init__(name="Draft_AddConstruction")

    def GetResources(self):
        """Set icon, menu and tooltip."""
        return {
            "Pixmap": "Draft_AddConstruction",
            "MenuText": QT_TRANSLATE_NOOP("Draft_AddConstruction", "Add to Construction Group"),
            "ToolTip": QT_TRANSLATE_NOOP(
                "Draft_AddConstruction",
                "Adds the selected objects to the construction group,\nand changes their appearance to the construction style.\nThe construction group is created if it does not exist.",
            ),
        }

    def Activated(self):
        """Execute when the command is called."""
        super().Activated()

        if not hasattr(Gui, "draftToolBar"):
            return

        self.doc.openTransaction(translate("draft", "Add to Construction Group"))
        col = params.get_param("constructioncolor") | 0x000000FF

        # Get the construction group or create it if it doesn't exist
        grp = self.doc.getObject("Draft_Construction")
        if not grp:
            grp = self.doc.addObject("App::DocumentObjectGroup", "Draft_Construction")
            grp.Label = params.get_param("constructiongroupname")

        for obj in Gui.Selection.getSelection():
            grp.addObject(obj)
            # Change the appearance to the construction colors
            vobj = obj.ViewObject
            if "TextColor" in vobj.PropertiesList:
                vobj.TextColor = col
            if "PointColor" in vobj.PropertiesList:
                vobj.PointColor = col
            if "LineColor" in vobj.PropertiesList:
                vobj.LineColor = col
            if "ShapeColor" in vobj.PropertiesList:
                vobj.ShapeColor = col
            if hasattr(vobj, "Transparency"):
                vobj.Transparency = 80

        self.doc.commitTransaction()
        self.doc.recompute()


Draft_AddConstruction = AddToConstruction
Gui.addCommand("Draft_AddConstruction", AddToConstruction())


class AddNamedGroup(gui_base.GuiCommandSimplest):
    """GuiCommand for the Draft_AddNamedGroup tool.

    It adds a new named group.
    """

    def __init__(self):
        super().__init__(name="Draft_AddNamedGroup")

    def GetResources(self):
        """Set icon, menu and tooltip."""
        return {
            "Pixmap": "Draft_AddNamedGroup",
            "MenuText": QT_TRANSLATE_NOOP("Draft_AddNamedGroup", "New Named Group"),
            "ToolTip": QT_TRANSLATE_NOOP("Draft_AddNamedGroup", "Adds a group with a given name"),
        }

    def Activated(self):
        super().Activated()

        txt, ok = QtWidgets.QInputDialog.getText(
            None,
            translate("draft", "New Group"),
            translate("draft", "Group name"),
            text=translate("draft", "Group", "Object label"),
        )
        if not ok:
            return
        if not txt:
            return
        self.doc.openTransaction(translate("draft", "New named group"))
        grp = self.doc.addObject("App::DocumentObjectGroup", translate("draft", "Group"))
        grp.Label = txt
        moveToGroup(grp)
        self.doc.commitTransaction()
        self.doc.recompute()


Gui.addCommand("Draft_AddNamedGroup", AddNamedGroup())

## @}



# SPDX-License-Identifier: LGPL-2.1-or-later

# ***************************************************************************
# *                                                                         *
# *   Copyright (c) 2019 Yorik van Havre <yorik@uncreated.net>              *
# *                                                                         *
# *   This file is part of FreeCAD.                                         *
# *                                                                         *
# *   FreeCAD is free software: you can redistribute it and/or modify it    *
# *   under the terms of the GNU Lesser General Public License as           *
# *   published by the Free Software Foundation, either version 2.1 of the  *
# *   License, or (at your option) any later version.                       *
# *                                                                         *
# *   FreeCAD is distributed in the hope that it will be useful, but        *
# *   WITHOUT ANY WARRANTY; without even the implied warranty of            *
# *   MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE. See the GNU      *
# *   Lesser General Public License for more details.                       *
# *                                                                         *
# *   You should have received a copy of the GNU Lesser General Public      *
# *   License along with FreeCAD. If not, see                               *
# *   <https://www.gnu.org/licenses/>.                                      *
# *                                                                         *
# ***************************************************************************

import os

import FreeCAD
import FreeCADGui

QT_TRANSLATE_NOOP = FreeCAD.Qt.QT_TRANSLATE_NOOP
translate = FreeCAD.Qt.translate


class BIM_IfcExplorer:

    def __init__(self):
        self.tree = None

    def GetResources(self):
        import Arch_rc

        return {
            "Pixmap": "IFC",
            "MenuText": QT_TRANSLATE_NOOP("BIM_IfcExplorer", "IFC Explorer"),
            "ToolTip": QT_TRANSLATE_NOOP("BIM_IfcExplorer", "Opens the IFC explorer utility"),
        }

    def Activated(self):

        from PySide import QtGui

        try:
            import ifcopenshell
        except ImportError:
            FreeCAD.Console.PrintError(
                translate(
                    "BIM",
                    "IfcOpenShell was not found on this system. IFC support is disabled",
                )
                + "\n"
            )
            return

        # setting up a font
        self.bold = QtGui.QFont()
        self.bold.setBold(True)

        # setting up a link fint
        self.linkfont = QtGui.QFont()
        self.linkfont.setUnderline(True)

        # setup a brush to paint text in system link color
        self.linkbrush = QtGui.QApplication.palette().link()

        # draw the main tree widget
        self.tree = QtGui.QTreeWidget()
        self.tree.setColumnCount(1)
        self.tree.setWordWrap(True)
        self.tree.header().setDefaultSectionSize(60)
        self.tree.header().resizeSection(0, 180)
        self.tree.header().setStretchLastSection(True)
        self.tree.headerItem().setText(0, translate("BIM", "Objects structure"))

        # draw the attributes widget
        self.attributes = QtGui.QTreeWidget()
        self.attributes.setColumnCount(2)
        self.attributes.setWordWrap(True)
        self.attributes.header().setDefaultSectionSize(60)
        self.attributes.header().resizeSection(0, 120)
        self.attributes.header().resizeSection(1, 200)
        self.attributes.header().setStretchLastSection(True)
        self.attributes.headerItem().setText(0, translate("BIM", "Attribute"))
        self.attributes.headerItem().setText(1, translate("BIM", "Value"))

        # draw the properties widget
        self.properties = QtGui.QTreeWidget()
        self.properties.setColumnCount(2)
        self.properties.setWordWrap(True)
        self.properties.header().setDefaultSectionSize(60)
        self.properties.header().resizeSection(0, 120)
        self.properties.header().resizeSection(1, 200)
        self.properties.header().setStretchLastSection(True)
        self.properties.headerItem().setText(0, translate("BIM", "Property"))
        self.properties.headerItem().setText(1, translate("BIM", "Value"))

        # create the dialog
        self.dialog = QtGui.QDialog()
        self.dialog.setObjectName("IfcExplorer")
        self.dialog.setWindowTitle(translate("BIM", "IFC Explorer"))
        self.dialog.resize(720, 540)
        toolbar = FreeCADGui.UiLoader().createWidget("Gui::ToolBar")

        layout = QtGui.QVBoxLayout()
        layout.addWidget(toolbar)
        hlayout = QtGui.QHBoxLayout()
        hlayout.addWidget(self.tree)
        layout.addLayout(hlayout)
        vlayout = QtGui.QVBoxLayout()
        hlayout.addLayout(vlayout)
        vlayout.addWidget(self.attributes)
        vlayout.addWidget(self.properties)
        self.dialog.setLayout(layout)

        # draw the toolbar buttons
        self.openAction = QtGui.QAction(translate("BIM", "Open"), None)
        self.openAction.setToolTip(translate("BIM", "Open another IFC file"))
        self.openAction.triggered.connect(self.open)
        self.openAction.setIcon(QtGui.QIcon(":/icons/document-open.svg"))
        toolbar.addAction(self.openAction)

        self.backAction = QtGui.QAction(translate("BIM", "Back"), None)
        self.backAction.setToolTip(translate("BIM", "Go back to last item selected"))
        self.backAction.triggered.connect(self.back)
        self.backAction.setIcon(QtGui.QIcon(":/icons/edit-undo.svg"))
        toolbar.addAction(self.backAction)

        self.shapeAction = QtGui.QAction(translate("BIM", "Insert"), None)
        self.shapeAction.setToolTip(
            translate(
                "BIM",
                "Inserts the selected object and its children in the active document",
            )
        )
        self.shapeAction.triggered.connect(self.insert)
        self.shapeAction.setIcon(QtGui.QIcon(":icons/Tree_Part.svg"))
        self.shapeAction.setEnabled(False)
        toolbar.addAction(self.shapeAction)

        self.meshAction = QtGui.QAction(translate("BIM", "Mesh"), None)
        self.meshAction.setToolTip(translate("BIM", "Turn mesh display on/off"))
        self.meshAction.triggered.connect(self.toggleMesh)
        self.meshAction.setCheckable(True)
        self.meshAction.setChecked(False)
        self.meshAction.setIcon(QtGui.QIcon(":/icons/DrawStyleShaded.svg"))
        toolbar.addAction(self.meshAction)

        # connect signals/slots
        self.tree.currentItemChanged.connect(self.onSelectTree)
        self.attributes.itemDoubleClicked.connect(self.onDoubleClickTree)
        self.properties.itemDoubleClicked.connect(self.onDoubleClickTree)
        self.dialog.rejected.connect(self.close)

        # center the dialog over FreeCAD window
        mw = FreeCADGui.getMainWindow()
        self.dialog.move(
            mw.frameGeometry().topLeft() + mw.rect().center() - self.dialog.rect().center()
        )

        # open a file and show the dialog
        self.open()
        if self.filename:
            self.dialog.show()

    def open(self):
        "opens a file"

        import ifcopenshell
        from PySide import QtGui

        self.filename = ""
        lastfolder = FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/BIM").GetString(
            "lastIfcExplorerFolder", ""
        )
        filename = QtGui.QFileDialog.getOpenFileName(
            None,
            translate("BIM", "Select an IFC file"),
            lastfolder,
            translate("BIM", "IFC files (*.ifc)"),
        )
        if filename and filename[0]:
            self.filename = filename[0]
            FreeCAD.ParamGet("User parameter:BaseApp/Preferences/Mod/BIM").SetString(
                "lastIfcExplorerFolder", os.path.dirname(self.filename)
            )
            if not os.path.exists(self.filename):
                FreeCAD.Console.PrintError(translate("BIM", "File not found") + "\n")
                return
        else:
            return

        # set window title
        self.dialog.setWindowTitle(
            translate("BIM", "IFC Explorer") + " - " + os.path.basename(self.filename)
        )

        # clear everything
        self.tree.clear()
        self.attributes.clear()
        self.properties.clear()
        self.done = []
        self.backnav = []
        self.mesh = None
        self.products = []
        self.omeshes = {}
        self.currentmesh = None

        # read file and order contents
        self.ifc = ifcopenshell.open(self.filename)
        root = self.getEntitiesTree()

        # unable to find IfcSite
        if not root:
            FreeCAD.Console.PrintError(
                translate(
                    "BIM",
                    "IfcSite element was not found in %s. Unable to explore.",
                )
                % self.filename
                + "\n"
            )
            return

        # populate tree contents
        for eid, children in root.items():
            self.addEntity(eid, children, self.tree)
        # self.tree.expandAll()

    def close(self):
        "close the dialog"

        if FreeCAD.ActiveDocument:
            if getattr(self, "mesh", None):
                FreeCAD.ActiveDocument.removeObject(self.mesh.Name)
            if getattr(self, "currentmesh", None):
                FreeCAD.ActiveDocument.removeObject(self.currentmesh.Name)

    def back(self):
        "selects the previously selected item in the tree"

        if self.backnav:
            item = self.backnav.pop()
            self.tree.setCurrentItem(item)

    def insert(self):
        "inserts selected objects in the active document"

        from importers import importIFC
        from PySide import QtCore

        doc = FreeCAD.ActiveDocument
        if doc and self.filename:
            item = self.tree.currentItem()
            if item:
                eid = item.data(0, QtCore.Qt.UserRole)
                if eid:
                    importIFC.ZOOMOUT = False
                    try:
                        importIFC.insert(self.ifc, doc.Name, only=[eid])
                    except TypeError:
                        importIFC.insert(self.filename, doc.Name, only=[eid])
                    if self.currentmesh:
                        self.currentmesh.ViewObject.hide()

    def toggleMesh(self, checked=False):
        "turns mesh display on/off"

        import Mesh
        import ifcopenshell
        from ifcopenshell import geom

        if not FreeCAD.ActiveDocument:
            doc = FreeCAD.newDocument()
            FreeCAD.setActiveDocument(doc.Name)
        if FreeCAD.ActiveDocument:
            if checked:
                if self.mesh:
                    self.mesh.ViewObject.show()
                else:
                    try:
                        from importers import importIFCHelper

                        s = importIFCHelper.getScaling(self.ifc)
                    except:
                        from importers import importIFC

                        s = importIFC.getScaling(self.ifc)
                    s *= 1000  # ifcopenshell outputs its meshes in metres
                    trf = None
                    if s != 1:
                        trf = FreeCAD.Matrix()
                        trf.scale(s, s, s)
                    basemesh = Mesh.Mesh()
                    s = geom.settings()
                    s.set(s.USE_WORLD_COORDS, True)
                    for product in self.products:
                        try:
                            m = geom.create_shape(s, product)
                            g = m.geometry
                            v = g.verts
                            f = g.faces
                            verts = [FreeCAD.Vector(v[i : i + 3]) for i in range(0, len(v), 3)]
                            faces = [tuple(f[i : i + 3]) for i in range(0, len(f), 3)]
                            omesh = Mesh.Mesh((verts, faces))
                            if trf:
                                omesh.transform(trf)
                            self.omeshes[product.id()] = omesh
                            basemesh.addMesh(omesh)
                        except:
                            pass
                    self.mesh = FreeCAD.ActiveDocument.addObject("Mesh::Feature", "IFCMesh")
                    self.mesh.Mesh = basemesh
                    self.mesh.ViewObject.Transparency = 85
                    FreeCAD.ActiveDocument.recompute()
                    FreeCADGui.Selection.clearSelection()
                    FreeCADGui.Selection.addSelection(self.mesh)
                    FreeCADGui.SendMsgToActiveView("ViewSelection")
            else:
                if self.mesh:
                    self.mesh.ViewObject.hide()
                if self.currentmesh:
                    self.currentmesh.ViewObject.hide()

    def getEntitiesTree(self):
        "stacks the entities inside containers"

        root = {}

        for site in self.ifc.by_type("IfcSite"):
            root[site.id()] = self.getChildren(site)

        return root

        # entities =  ifc.by_type("IfcRoot")
        # entities += ifc.by_type("IfcRepresentation")
        # entities += ifc.by_type("IfcRepresentationItem")
        # entities += ifc.by_type("IfcRepresentationMap")
        # entities += ifc.by_type("IfcPlacement")
        # entities += ifc.by_type("IfcProperty")
        # entities += ifc.by_type("IfcPhysicalSimpleQuantity")
        # entities += ifc.by_type("IfcMaterial")
        # entities += ifc.by_type("IfcProductRepresentation")
        # entities = sorted(entities, key=lambda eid: eid.id())

    def getChildren(self, obj, keys=False):
        "returns a recursive dict of the children of this obj"

        children = {}
        if obj.is_a("IfcProduct"):
            self.products.append(obj)
        if hasattr(obj, "IsDecomposedBy"):  # building structure
            for rel in obj.IsDecomposedBy:
                if hasattr(rel, "RelatedObjects"):
                    for child in rel.RelatedObjects:
                        children[child.id()] = self.getChildren(child)
        if hasattr(obj, "ContainsElements"):  # objects inside building structure
            for rel in obj.ContainsElements:
                if hasattr(rel, "RelatedElements"):
                    for child in rel.RelatedElements:
                        children[child.id()] = self.getChildren(child)
        if hasattr(obj, "Representation"):  # Shape representation
            if obj.Representation:
                children[obj.Representation.id()] = self.getChildren(obj.Representation)
        if hasattr(obj, "Representations"):
            for rep in obj.Representations:
                children[rep.id()] = self.getChildren(rep)
        if obj.is_a("IfcShapeRepresentation"):
            for it in obj.Items:
                children[it.id()] = self.getChildren(it)
        if keys:

            def getkeys(d):
                ck = list(d.keys())
                for v in d.values():
                    ck.extend(getkeys(v))
                return ck

            return getkeys(children)
        return children

    def addEntity(self, eid, children, parent):
        "adds a given entity and its children to the given tree item"

        from PySide import QtCore, QtGui

        def get_name(e):
            try:
                return " : " + e.get_info()["Name"]
            except:
                return ""

        if not eid in self.done:
            entity = self.ifc[eid]
            item = QtGui.QTreeWidgetItem(parent)
            # item.setText(0,self.tostr(eid))
            name = ""
            if entity.is_a("IfcProduct"):
                name = get_name(entity)
                item.setFont(0, self.bold)
                if isinstance(parent, QtGui.QTreeWidgetItem):
                    parent.setExpanded(True)
            item.setText(0, "#" + self.tostr(eid) + " : " + self.tostr(entity.is_a()) + name)
            if entity.is_a() in ["IfcWall", "IfcWallStandardCase"]:
                item.setIcon(0, QtGui.QIcon(":icons/Arch_Wall_Tree.svg"))
            elif entity.is_a() in ["IfcBuildingElementProxy"]:
                item.setIcon(0, QtGui.QIcon(":icons/Arch_Component.svg"))
            elif entity.is_a() in [
                "IfcColumn",
                "IfcColumnStandardCase",
                "IfcBeam",
                "IfcBeamStandardCase",
                "IfcSlab",
                "IfcFooting",
                "IfcPile",
                "IfcTendon",
            ]:
                item.setIcon(0, QtGui.QIcon(":icons/Arch_Structure_Tree.svg"))
            elif entity.is_a() in ["IfcSite"]:
                item.setIcon(0, QtGui.QIcon(":icons/Arch_Site_Tree.svg"))
            elif entity.is_a() in ["IfcBuilding"]:
                item.setIcon(0, QtGui.QIcon(":icons/Arch_Building_Tree.svg"))
            elif entity.is_a() in ["IfcBuildingStorey"]:
                item.setIcon(0, QtGui.QIcon(":icons/Arch_Floor_Tree.svg"))
            elif entity.is_a() in [
                "IfcWindow",
                "IfcWindowStandardCase",
                "IfcDoor",
                "IfcDoorStandardCase",
            ]:
                item.setIcon(0, QtGui.QIcon(":icons/Arch_Window_Tree.svg"))
            elif entity.is_a() in ["IfcRoof"]:
                item.setIcon(0, QtGui.QIcon(":icons/Arch_Roof_Tree.svg"))
            elif entity.is_a() in ["IfcExtrudedAreaSolid", "IfcClosedShell"]:
                item.setIcon(0, QtGui.QIcon(":icons/Tree_Part.svg"))
            elif entity.is_a() in ["IfcFace"]:
                item.setIcon(0, QtGui.QIcon(":icons/Draft_SwitchMode.svg"))
            elif entity.is_a() in ["IfcArbitraryClosedProfileDef", "IfcPolyloop"]:
                item.setIcon(0, QtGui.QIcon(":icons/Draft_Draft.svg"))
            elif entity.is_a() in [
                "IfcPropertySingleValue",
                "IfcQuantityArea",
                "IfcQuantityVolume",
            ]:
                item.setIcon(0, QtGui.QIcon(":icons/Tree_Annotation.svg"))
            elif entity.is_a() in ["IfcMaterial"]:
                item.setIcon(0, QtGui.QIcon(":icons/Arch_Material.svg"))
            elif entity.is_a() in ["IfcReinforcingBar"]:
                item.setIcon(0, QtGui.QIcon(":icons/Arch_Rebar.svg"))
            elif entity.is_a("IfcProduct"):
                item.setIcon(0, QtGui.QIcon(":icons/Arch_Component.svg"))
            item.setFirstColumnSpanned(True)
            item.setData(0, QtCore.Qt.UserRole, eid)
            for childid, grandchildren in children.items():
                self.addEntity(childid, grandchildren, item)
            self.done.append(eid)

    def addAttributes(self, eid, parent):
        "adds the attributes of the given IFC entity under the given QTreeWidgetITem"

        import ifcopenshell
        from PySide import QtGui

        entity = self.ifc[eid]

        i = 0
        while True:
            try:
                argname = entity.attribute_name(i)
            except RuntimeError:
                break
            else:
                try:
                    argvalue = getattr(entity, argname)
                except AttributeError:
                    FreeCAD.Console.PrintError(
                        translate("BIM", "Error in entity") + " " + self.tostr(entity) + "\n"
                    )
                    break
                else:
                    if argname not in ["Id", "GlobalId"]:
                        colored = False
                        if isinstance(argvalue, ifcopenshell.entity_instance):
                            if argvalue.id() == 0:
                                t = self.tostr(argvalue)
                            else:
                                colored = True
                                t = (
                                    "#"
                                    + self.tostr(argvalue.id())
                                    + ": "
                                    + self.tostr(argvalue.is_a())
                                )
                        elif isinstance(argvalue, (list, tuple)):
                            t = ""
                        else:
                            t = self.tostr(argvalue)
                        item = QtGui.QTreeWidgetItem(parent)
                        item.setText(0, self.tostr(argname))
                        if t and (t != "None"):
                            item.setText(1, t)
                            if colored:
                                item.setForeground(1, self.linkbrush)
                                item.setFont(1, self.linkfont)
                            if argname == "Name":
                                item.setFont(1, self.bold)
                        if isinstance(argvalue, (list, tuple)):
                            j = 0
                            for argitem in argvalue:
                                colored = False
                                if isinstance(argitem, ifcopenshell.entity_instance):
                                    if argitem.id() == 0:
                                        t = self.tostr(argitem)
                                    else:
                                        colored = True
                                        t = (
                                            "#"
                                            + self.tostr(argitem.id())
                                            + ": "
                                            + self.tostr(argitem.is_a())
                                        )
                                else:
                                    t = argitem
                                t = self.tostr(t)
                                if j == 0:
                                    item.setText(1, t)
                                    if colored:
                                        item.setForeground(1, self.linkbrush)
                                        item.setFont(1, self.linkfont)
                                else:
                                    subitem = QtGui.QTreeWidgetItem(item)
                                    subitem.setText(1, t)
                                    if colored:
                                        subitem.setForeground(1, self.linkbrush)
                                        subitem.setFont(1, self.linkfont)
                                j += 1
                i += 1

    def addProperties(self, eid, parent):
        "adds properties of a given entity to the given QTReeWidgetItem"

        from PySide import QtGui

        entity = self.ifc[eid]
        if hasattr(entity, "IsDefinedBy"):
            for rel in entity.IsDefinedBy:
                if hasattr(rel, "RelatingPropertyDefinition"):
                    if rel.RelatingPropertyDefinition:
                        item = QtGui.QTreeWidgetItem(parent)
                        item.setText(
                            0,
                            "PropertySet: " + self.tostr(rel.RelatingPropertyDefinition.Name),
                        )
                        item.setFont(0, self.bold)
                        item.setFirstColumnSpanned(True)
                        if hasattr(rel.RelatingPropertyDefinition, "HasProperties"):
                            for prop in rel.RelatingPropertyDefinition.HasProperties:
                                subitem = QtGui.QTreeWidgetItem(item)
                                subitem.setText(0, "Property")
                                self.addAttributes(prop.id(), subitem)

    def tostr(self, text):
        if isinstance(text, str):
            return text
        else:
            return str(text)

    def onSelectTree(self, item, previous):
        "displays attributes and properties of a tree item"

        from PySide import QtCore

        self.backnav.append(previous)
        eid = item.data(0, QtCore.Qt.UserRole)
        self.attributes.clear()
        self.addAttributes(eid, self.attributes)
        self.attributes.expandAll()
        self.properties.clear()
        self.addProperties(eid, self.properties)
        self.properties.expandAll()
        entity = self.ifc[eid]
        if entity.is_a("IfcProduct") and FreeCAD.ActiveDocument:
            self.shapeAction.setEnabled(True)
        else:
            self.shapeAction.setEnabled(False)
        if eid in self.omeshes:
            omesh = self.omeshes[eid]
        else:
            omesh = None
        children = self.getChildren(entity, keys=True)
        for k in children:
            if k in self.omeshes:
                kmesh = self.omeshes[k]
                if omesh:
                    omesh.addMesh(kmesh)
                else:
                    omesh = kmesh
        if omesh:
            if not self.currentmesh:
                self.currentmesh = FreeCAD.ActiveDocument.addObject(
                    "Mesh::Feature", "IFCObjectMesh"
                )
                self.currentmesh.ViewObject.ShapeColor = (0.4, 0.4, 1.0)
            self.currentmesh.Mesh = omesh
            self.currentmesh.ViewObject.show()
            FreeCAD.ActiveDocument.recompute()
        else:
            if self.currentmesh:
                self.currentmesh.ViewObject.hide()

    def onDoubleClickTree(self, item, column):
        "when a property or attribute is double-clicked"

        from PySide import QtCore

        if self.tree:
            txt = item.text(column)
            if txt.startswith("#"):
                eid = txt[1:].split(":")[0]
                target = self.tree.findItems(
                    "#" + eid + " :",
                    QtCore.Qt.MatchStartsWith | QtCore.Qt.MatchRecursive,
                    0,
                )
                if target:
                    self.tree.scrollToItem(target[0])
                    self.tree.setCurrentItem(target[0])


FreeCADGui.addCommand("BIM_IfcExplorer", BIM_IfcExplorer())



# SPDX-License-Identifier: LGPL-2.1-or-later
# SPDX-FileCopyrightText: 2025 FreeCAD Project Association
# SPDX-FileNotice: Part of the AddonManager.

################################################################################
#                                                                              #
#   This addon is free software: you can redistribute it and/or modify         #
#   it under the terms of the GNU Lesser General Public License as             #
#   published by the Free Software Foundation, either version 2.1              #
#   of the License, or (at your option) any later version.                     #
#                                                                              #
#   This addon is distributed in the hope that it will be useful,              #
#   but WITHOUT ANY WARRANTY; without even the implied warranty                #
#   of MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                    #
#   See the GNU Lesser General Public License for more details.                #
#                                                                              #
#   You should have received a copy of the GNU Lesser General Public           #
#   License along with this addon. If not, see https://www.gnu.org/licenses    #
#                                                                              #
################################################################################

from enum import Enum
from PySideWrapper import QtWidgets
from addonmanager_utilities import get_main_am_window


class MessageDialog:

    class DialogType(Enum):
        INFO = 0
        WARNING = 1
        ERROR = 2
        QUESTION = 3

    @staticmethod
    def show_modal(
        dialog_type: DialogType,
        object_name: str,
        title: str,
        message: str,
        button_set: QtWidgets.QDialogButtonBox.StandardButton = QtWidgets.QDialogButtonBox.StandardButton.Ok,
        parent=None,
    ):
        """This method works the same as things like `QMessageBox.critical` and friends, displaying
        a modal dialog with the given type, title and message. By default, the OK button is used,
        but a different set of buttons can be specified. The main difference between this and the
        Qt version is that the dialog's objectName is set, allowing it to be found by name later,
        especially during unit testing.

        :param dialog_type: The type of dialog to show (INFO, WARNING, ERROR, QUESTION)
        :param object_name: The name of the dialog object, passed to setObjectName(). DO NOT TRANSLATE.
        :param title: The title of the dialog (should be translated)
        :param message: The message to display (should be translated)
        :param button_set: The set of buttons to display (defaults to OK)
        :param parent: The parent widget for the dialog (defaults to the main window of the Addon Manager)
        """
        dialog = QtWidgets.QMessageBox(parent or get_main_am_window())
        dialog.setObjectName(object_name)
        dialog.setWindowTitle(title)
        dialog.setText(message)
        if dialog_type == MessageDialog.DialogType.QUESTION:
            dialog.setIcon(QtWidgets.QMessageBox.Icon.Question)
        elif dialog_type == MessageDialog.DialogType.WARNING:
            dialog.setIcon(QtWidgets.QMessageBox.Icon.Warning)
        elif dialog_type == MessageDialog.DialogType.ERROR:
            dialog.setIcon(QtWidgets.QMessageBox.Icon.Critical)
        else:
            dialog.setIcon(QtWidgets.QMessageBox.Icon.Information)
        dialog.setStandardButtons(button_set)
        return dialog.exec()
